#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "../include/string.h"
#include "obj.h"
#include "messages.h"
#include "mission.h"
#include "simutils.h"
#include "sar.h"


int SARMissionMatchNameInContact(
	sar_core_struct *core_ptr,
	sar_object_struct *obj_ptr,
	const char *name
);
char *SARMissionFormatDeltaTime(time_t t);

void SARMissionPrintStats(
	sar_core_struct *core_ptr,
	sar_scene_struct *scene,
        sar_mission_struct *mission,
        sar_object_struct *obj_ptr
);

void SARMissionLandNotifyArrive(
	sar_core_struct *core_ptr,
	sar_mission_struct *mission,
        sar_object_struct *obj_ptr
);
void SARMissionLandNotifyRescue(
	sar_core_struct *core_ptr,
	sar_mission_struct *mission,
	sar_object_struct *obj_ptr
);

void SARMissionManageArrive(
	sar_core_struct *core_ptr, sar_mission_struct *mission
);
void SARMissionManageRescue(
	sar_core_struct *core_ptr, sar_mission_struct *mission
);

void SARMissionDestroyNotify(
	sar_core_struct *core_ptr, sar_object_struct *obj_ptr
);
void SARMissionLandNotify(
	sar_core_struct *core_ptr, sar_object_struct *obj_ptr
);
int SARMissionManage(sar_core_struct *core_ptr);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))


/*
 * 	Interval between mission end check and going back to
 *	menus.
 */
#define DEF_MISSION_END_INT	5000

/*
 *	Checks if object is within the contact radius (cylendrical,
 *	infinate z) with the object specified by name. Returns
 *	the matched object number or -1 on failure.
 */
int SARMissionMatchNameInContact(
	sar_core_struct *core_ptr,
	sar_object_struct *obj_ptr,
        const char *name
)
{
	int i;
	sar_object_struct	*tar_obj_ptr,
				*src_obj_ptr = obj_ptr;
	sar_position_struct *pos_tar, *pos_src;
	double d, tar_contact_radius;


	if((core_ptr == NULL) ||
	   (src_obj_ptr == NULL) ||
           (name == NULL)
	)
	    return(-1);

	pos_src = &src_obj_ptr->pos;

	/* Go through list of objects on core structure. */
	for(i = 0; i < core_ptr->total_objects; i++)
	{
	    tar_obj_ptr = core_ptr->object[i];
	    if(tar_obj_ptr == NULL)
		continue;

	    /* Skip source object. */
	    if(src_obj_ptr == tar_obj_ptr)
		continue;

	    pos_tar = &tar_obj_ptr->pos;

	    /* Calculate distance apart. */
	    d = hypot(
		pos_tar->x - pos_src->x,
		pos_tar->y - pos_src->y
	    );

	    /* Get 2d contact radius for the target object. */
	    tar_contact_radius = SARSimGetFlatContactRadius(
		tar_obj_ptr
	    );
	    if(d > tar_contact_radius)
		continue;

	    /* Object names match? */
	    if(tar_obj_ptr->name != NULL)
	    {
		if(!strcasecmp(tar_obj_ptr->name, name))
		    break;
	    }
	}
	if(i < core_ptr->total_objects)
	    return(i);
	else
	    return(-1);
}


/*
 *	Returns a statically allocated string verbosly stating the
 *	delta time t.
 *
 *	Note that this function never returns NULL.
 */
char *SARMissionFormatDeltaTime(time_t t)
{
#define len 256
	static char s[len];


	if(t <= 0)
	    return("none");

	if(t < 60)
	{
	    sprintf(s, "%ld second%s",
		t,
		(t == 1) ? "" : "s"
	    );
	}
	else if(t < 3600)
	{
	    t = (time_t)(t / 60);
	    sprintf(s, "%ld minute%s",
                t,
                (t == 1) ? "" : "s"
            );
        }
	else if(t < 86400)
	{
            t = (time_t)(t / 3600);
            sprintf(s, "%ld hour%s",
                t,
                (t == 1) ? "" : "s"
            );
        }

	return(s);
#undef len
}

/*
 *	Prints mission stats to the scene's message output with respect
 *	to the given object and mission structures.
 */
void SARMissionPrintStats(
	sar_core_struct *core_ptr, sar_scene_struct *scene,
        sar_mission_struct *mission,
        sar_object_struct *obj_ptr
)
{
	int	passengers = 0, 	/* Current on board. */
		passengers_max = 0;	/* Maximum capacity. */
	sar_object_aircraft_struct *obj_aircraft_ptr;
	char text[256];


	/* First print object statistics if object is valid. */
	if(obj_ptr != NULL)
	{
	    /* Handle by object type. */
	    switch(obj_ptr->type)
	    {
	      case SAR_OBJ_TYPE_AIRCRAFT:
		obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
		if(obj_aircraft_ptr != NULL)
		{
		    passengers = obj_aircraft_ptr->passengers;
		    passengers_max = obj_aircraft_ptr->passengers_max;
		}
		break;
	    }

	    /* Format passengers on board string. */
	    sprintf(
		text,
		"Passengers: %i(%i)",
		passengers,
		passengers_max
	    );
	    SARMessageAdd(scene, text);
	}

	/* Mission status. */
	if(mission != NULL)
	{
	    sar_mobj_arrive_struct *arrive_objective;
	    sar_mobj_rescue_struct *rescue_objective;
#define tmp_name_len	80
	    char tmp_name[tmp_name_len];


            /* Handle by mission type. */
            switch(mission->type)
            {
              case MISSION_TYPE_ARRIVE:
		arrive_objective = mission->objective;
		if(arrive_objective == NULL)
		    break;

		strncpy(
		    tmp_name,
		    ((arrive_objective->arrive_at_name == NULL) ?
			"unknown" : arrive_objective->arrive_at_name
		    ),
		    tmp_name_len
		);
		tmp_name[tmp_name_len - 1] = '\0';

		/* Print arrive mission status. */
		switch(mission->state)
		{
		  case MISSION_STATE_IN_PROGRESS:
		    sprintf(text,
			"Mission in progress, enroute to %s",
			tmp_name
		    );
		    SARMessageAdd(scene, text);
		    if(arrive_objective->time_left > 0.0)
		    {
			sprintf(text,
                            "Time left %s",
                            SARMissionFormatDeltaTime(
				(time_t)arrive_objective->time_left
			    )
                        );
                        SARMessageAdd(scene, text);
		    }
		    break;

		  case MISSION_STATE_ACCOMPLISHED:
                    sprintf(text, "Mission objective achieved!");
		    SARMessageAdd(scene, text);
                    break;

                  case MISSION_STATE_FAILED:
                    sprintf(text, "Mission objective failed.");
                    SARMessageAdd(scene, text);
                    break;
		}
		break;

	      case MISSION_TYPE_RESCUE:
                rescue_objective = mission->objective;
                if(rescue_objective == NULL)
                    break;

		/* Print rescue mission status. */
                switch(mission->state)
                {
                  case MISSION_STATE_IN_PROGRESS:
		    /* Print how many more to find, remember to
		     * calculate this by; how_need_rescue -
		     * already_rescued - those_onboard
		     */
                    sprintf(
			text,
                        "Rescue in progress, %i more to find",
                        MAX(
			    rescue_objective->humans_need_rescue -
			    rescue_objective->humans_rescued -
			    passengers,
			    0
			)
                    );
                    SARMessageAdd(scene, text);
                    if(rescue_objective->time_left > 0.0)
                    {
                        sprintf(text,
                            "Time left %s",
                            SARMissionFormatDeltaTime(
                                (time_t)rescue_objective->time_left
                            )
                        );
                        SARMessageAdd(scene, text);
                    }
                    break;

                  case MISSION_STATE_ACCOMPLISHED:
                    sprintf(text, "Mission objective achieved!");
                    SARMessageAdd(scene, text);
                    break;

                  case MISSION_STATE_FAILED:
                    sprintf(text, "Mission objective failed.");
                    SARMessageAdd(scene, text);
                    break;
                }
		break;
	    }
#undef tmp_name_len
        }

	return;
}


/*
 *	Handle land notify arrive mission.
 */
void SARMissionLandNotifyArrive(
	sar_core_struct *core_ptr,
	sar_mission_struct *mission,
        sar_object_struct *obj_ptr
)
{
        void *ptr = mission->objective;
        sar_mobj_arrive_struct *objective_ptr;
	sar_scene_struct *scene = core_ptr->scene;


	if(scene == NULL)
	    return;

        if((*(int *)ptr) != MISSION_TYPE_ARRIVE)
            return;
        else
            objective_ptr = ptr;

        /* Landed at right place? */
        if(SARMissionMatchNameInContact(
            core_ptr,
            obj_ptr,
            objective_ptr->arrive_at_name
        ) < 0)
            return;

        /* Mission accomplished? */
        if((mission->state == MISSION_STATE_IN_PROGRESS) &&
           (scene->player_obj_ptr == obj_ptr)
        )
        {
	    char *strptr = objective_ptr->message_success;

            mission->state = MISSION_STATE_ACCOMPLISHED;
            SARMessageAdd(
                scene,
		((strptr != NULL) ? strptr :
		    "Arrive mission complete."
		)
            );

	    /* Schedual next mission check in 5 seconds which will
	     * detect acomplished state.
	     */
	    mission->next_check = cur_millitime + DEF_MISSION_END_INT;
        }

	return;
}

/*
 *	Handle land notify rescue mission:
 */
void SARMissionLandNotifyRescue(
	sar_core_struct *core_ptr,
	sar_mission_struct *mission,
        sar_object_struct *obj_ptr
)
{
	int passengers_onboard = 0;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	void *ptr = mission->objective;
        sar_mobj_rescue_struct *objective_ptr;


        if((*(int *)ptr) != MISSION_TYPE_RESCUE)
            return;
        else
            objective_ptr = ptr;


	/* Landed at right place? */
	if(SARMissionMatchNameInContact(
	    core_ptr,
	    obj_ptr,
	    objective_ptr->arrive_at_name
	) < 0)
	    return;

	/* Get passengers on board. */
	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
	    obj_aircraft_ptr = obj_ptr->data;
	    if(obj_aircraft_ptr == NULL)
		break;

	    passengers_onboard = obj_aircraft_ptr->passengers;

	    /* Unload passengers. */
	    obj_aircraft_ptr->passengers = 0;
	    break;
	}

	/* Add humans rescued to mission objective. */
	objective_ptr->humans_rescued += passengers_onboard;

	/* Mission accomplished (do not check if this is the
	 * player object)?
	 */
	if((objective_ptr->humans_rescued >=
		objective_ptr->humans_need_rescue) &&
	   (mission->state == MISSION_STATE_IN_PROGRESS)
	)
	{
	    char *strptr = objective_ptr->message_success;

	    mission->state = MISSION_STATE_ACCOMPLISHED;
            SARMessageAdd(
                core_ptr->scene,
		((strptr != NULL) ? strptr :
		    "Rescue mission complete."
		)
            );

            /* Schedual next mission check in 5 seconds which will
             * detect acomplished state.
             */
            mission->next_check = cur_millitime + DEF_MISSION_END_INT;
	}

	return;
}

/*
 *	Manage arrive at mission.
 */
void SARMissionManageArrive(
        sar_core_struct *core_ptr,
        sar_mission_struct *mission
)
{
	void *ptr = mission->objective;
        sar_mobj_arrive_struct *objective_ptr;


	if(ptr == NULL)
	    return;

        if((*(int *)ptr) != MISSION_TYPE_ARRIVE)
            return;
        else
            objective_ptr = ptr;

	/* Time exceeded? */
	if(objective_ptr->time_left > 0.0)
	{
	    objective_ptr->time_left -= (double)(
		(double)lapsed_millitime * time_compression / 1000
	    );

	    if(objective_ptr->time_left <= 0.0)
	    {
		char *strptr = objective_ptr->message_fail;

		mission->state = MISSION_STATE_FAILED;
		SARMessageAdd(
		    core_ptr->scene,
		    ((strptr != NULL) ? strptr :
			"Arrive mission failed, time ran out."
		    )
		);

                /* Schedual next mission check in 5 seconds which will
                 * detect failed state.
                 */
                mission->next_check = cur_millitime + DEF_MISSION_END_INT;
	    }
	}

	return;
}

/*
 *	Manage rescue mission.
 */
void SARMissionManageRescue(
	sar_core_struct *core_ptr,
	sar_mission_struct *mission
)
{
        void *ptr = mission->objective;
        sar_mobj_rescue_struct *objective_ptr;

        
        if(ptr == NULL)
            return;

	if((*(int *)ptr) != MISSION_TYPE_RESCUE)
	    return;
	else
	    objective_ptr = ptr;

        /* Time exceeded? */
        if(objective_ptr->time_left > 0.0)
        {
            objective_ptr->time_left -= (double)(
		(double)lapsed_millitime * time_compression / 1000
            );

            if(objective_ptr->time_left <= 0.0)
            {
		char *strptr = objective_ptr->message_fail;

                mission->state = MISSION_STATE_FAILED;
                SARMessageAdd(
                    core_ptr->scene,
		    ((strptr != NULL) ? strptr :
			"Arrive mission failed, time ran out."
		    )
                );

                /* Schedual next mission check in 5 seconds which will
                 * detect failed state.
                 */
                mission->next_check = cur_millitime + DEF_MISSION_END_INT;
            }
        }

	return;
}



/*
 *	Destroy notify. This function should be called whenever an
 *	object is *about* to be destroyed.
 */
void SARMissionDestroyNotify(
        sar_core_struct *core_ptr,
        sar_object_struct *obj_ptr
)
{
        sar_mission_struct *mission;
	sar_scene_struct *scene;


        if((core_ptr == NULL) ||
           (obj_ptr == NULL)
        )
            return;

        mission = core_ptr->mission;
        if(mission == NULL)
            return;

	scene = core_ptr->scene;
	if(scene == NULL)
	    return;

	/* Is this the player object? */
	if(scene->player_obj_ptr == obj_ptr)
	{
            mission->state = MISSION_STATE_FAILED;

            /* Schedual next mission check in 5 seconds which will
             * detect failed state.
             */
            mission->next_check = cur_millitime + DEF_MISSION_END_INT;
	}

	return;
}

/*
 *	Land notify, this function should be called whenever an object
 *	has landed.
 */
void SARMissionLandNotify(
        sar_core_struct *core_ptr,
        sar_object_struct *obj_ptr
)
{
	sar_mission_struct *mission;


	if((core_ptr == NULL) ||
           (obj_ptr == NULL)
	)
	    return;

	mission = core_ptr->mission;
	if(mission == NULL)
	    return;

	/* Handle by mission type. */
	switch(mission->type)
	{
	  case MISSION_TYPE_ARRIVE:
	    SARMissionLandNotifyArrive(core_ptr, mission, obj_ptr);
	    break;

          case MISSION_TYPE_RESCUE:
            SARMissionLandNotifyRescue(core_ptr, mission, obj_ptr);
            break;
	}

	return;
}

/*
 *	Manages mission, this function should be called once per loop.
 *
 *	Return codes are as follows:
 *
 *	-1	Error.
 *	0	Nothing happened or not time to manage.
 *	1	Mission end success.
 *	2	Mission end failure.
 */
int SARMissionManage(sar_core_struct *core_ptr)
{
        sar_mission_struct *mission;
        
           
        if(core_ptr == NULL)
	    return(-1);

	mission = core_ptr->mission;
	if(mission == NULL)
	    return(-1);

	/* Update time spent on mission. */
	mission->time_spent += (double)(
	    (double)lapsed_millitime * time_compression / 1000
	);

	/* Not time to manage mission yet? */
	if(mission->next_check > cur_millitime)
	    return(0);
	else
	    mission->next_check = cur_millitime + mission->check_int;

	/* Mission no longer in progress? */
	if(mission->state != MISSION_STATE_IN_PROGRESS)
	{
	    switch(mission->state)
	    {
	      case MISSION_STATE_ACCOMPLISHED:
		return(1);
		break;

	      case MISSION_STATE_FAILED:
		return(2);
		break;

	      default:
		return(-1);
		break;
	    }
	}

	/* Handle by mission type. */
	switch(mission->type)
	{
	  case MISSION_TYPE_ARRIVE:
	    SARMissionManageArrive(core_ptr, mission);
	    break;

	  case MISSION_TYPE_RESCUE:
	    SARMissionManageRescue(core_ptr, mission);
	    break;
	}

	return(0);
}
