/*
                    SAR Objects and Scene Structures

            All coordinates and vectors use left hand rule.

 */

#ifndef OBJ_H
#define OBJ_H

#include <sys/types.h>

#include "../include/os.h"

#include "sfm.h"
#include "v3dtex.h"

#include "sound.h"


/*
 *	Engine flags type:
 */
#define sar_obj_flags_t	unsigned long

/*
 *	Gradient animation counter type:
 */
#define sar_grad_anim_t	u_int16_t

/*
 *	DMS units type:
 */
#define sar_dms_t	double


/*
 *	Air worthy states:
 */
#define SAR_AIR_WORTHY_NOT_FLYABLE	0	/* Pile of junk. */
#define SAR_AIR_WORTHY_OUT_OF_CONTROL	1
#define SAR_AIR_WORTHY_FLYABLE		2

/*
 *	Engine state:
 */
#define SAR_ENGINE_STATE_OFF		0
#define SAR_ENGINE_STATE_INIT		1
#define SAR_ENGINE_STATE_ON		2


/*
 *	Color structure:
 */
typedef struct {

	double a, r, g, b;

} sar_color_struct;

/*
 *	Visual model structure:
 */
typedef struct {

#define SAR_VISUAL_MODEL_NOT_LOADED	0
#define SAR_VISUAL_MODEL_LOADING	1
#define SAR_VISUAL_MODEL_LOADED		2

	int load_state;	/* One of SAR_VISUAL_MODEL_*. */

	int ref_count;	/* Referance count. */

	char *filename;	/* Can be NULL. */
	char *name;	/* Can be NULL. */

	void *data;	/* Really a GLint which referances a GL list. */

} sar_visual_model_struct;

/*
 *	Position structure:
 */
typedef struct {

	double x, y, z;
	long sx, sy, sz;

} sar_position_struct;

/*
 *	Direction structure:
 */
typedef struct {

	/* In radians. */
	double heading, pitch, bank;

} sar_direction_struct;


/*
 *	Intercept coordinate structure:
 */
typedef struct {

	sar_obj_flags_t flags;
	double x, y, z;		/* Position of intercept in meters. Note
				 * that z value is intercept altitude in
				 * meters.
				 */
	double radius;		/* Size of intercept cylendrical in meters. */
	double urgency;		/* How fast must we get there, 0.0 to 1.0. */
	char *name;		/* Intercept label. */

} sar_obj_intercept_struct;

/*
 *	Light structure:
 */
typedef struct {

#define SAR_LIGHT_FLAG_ON		(1 << 0)
#define SAR_LIGHT_FLAG_STROBE		(1 << 1)
#define SAR_LIGHT_FLAG_ATTENUATE	(1 << 2)

	sar_obj_flags_t flags;

	sar_position_struct pos;
	sar_direction_struct dir;

	int radius;	/* In pixels. */

	sar_color_struct color;			/* Apperance color. */
	sar_color_struct attenuate_color;	/* Attenuate color. */

	/* On/off timmings, only for SAR_LIGHT_TYPE_STROBE. */
	time_t	next_on, int_on,
		next_off, int_off;

} sar_light_struct;


/*
 *	Physical contact and crash boundaries structure:
 *
 *	Note that any object can have this contact bounds, but will be
 *	have to follow some special rules if they do.
 *
 *	First if the object is a human or aircraft, it should have only
 *	the SAR_CRASH_FLAG_CRASH_OTHER flag set and it's contact
 *	shape should be SAR_CONTACT_SHAPE_CYLENDRICAL.
 *
 *	Buildings and other objects that you want to set
 *	SAR_CRASH_FLAG_CRASH_CAUSE can have any shape (recommend
 *	SAR_CONTACT_SHAPE_CYLENDRICAL or SAR_CONTACT_SHAPE_RECTANGULAR).
 *
 *	Now objects that you want to set landable or walkable (with
 *	crash_flags set to SAR_CRASH_FLAG_SUPPORT_SURFACE) may not be
 *	human or aircrafts (objects that themselves need to be checked
 *	if on top of something).
 *
 *	If an object is set SAR_CRASH_FLAG_SUPPORT_SURFACE then the object
 *	can only have its heading rotated.
 *
 *	Note that object's of type SAR_OBJ_TYPE_GROUND are always
 *	landable/walkable and their shape is always cylendrical.
 */
typedef struct {

        /* Crash detection flags, specifies if this object can collide
         * with other objects, if other objects can collide with
         * this object, and if this object produces a flat support
	 * surface that is landable/walkable (respectivly with defined
	 * flags).
	 *
	 * If no flags are set it can be considered a quick way
	 * (ie if(crash_flags == 0)) to determine this object has no
         * crash interaction with other objects.
         */
#define SAR_CRASH_FLAG_CRASH_OTHER      (1 << 0)
#define SAR_CRASH_FLAG_CRASH_CAUSE      (1 << 1)
#define SAR_CRASH_FLAG_SUPPORT_SURFACE	(1 << 2)
        sar_obj_flags_t crash_flags;

        /* This value used if and only if the SAR_CRASH_FLAG_CRASH_CAUSE
	 * flag is set in crash_flags. This value specifies what type of
         * crash it would be if some other object crashed into this
	 * object.
         */
#define SAR_CRASH_TYPE_OBSTRUCTION      0       /* General structure. */
#define SAR_CRASH_TYPE_GROUND           1
#define SAR_CRASH_TYPE_MOUNTAIN         2
#define SAR_GRASH_TYPE_BUILDING         3
#define SAR_GRASH_TYPE_AIRCRAFT         4
        int crash_type;

        /* Body contact shape, one of SAR_CONTACT_SHAPE_*. */
#define SAR_CONTACT_SHAPE_SPHERICAL     0   
#define SAR_CONTACT_SHAPE_CYLENDRICAL   1
#define SAR_CONTACT_SHAPE_RECTANGULAR	2
        int contact_shape;

        /* Radius value in meters for contact shapes
	 * SAR_CONTACT_SHAPE_SPHERICAL or SAR_CONTACT_SHAPE_CYLENDRICAL.
	 */
        double contact_radius;

        /* Cylendrical height value in meters for contact shape
	 * SAR_CONTACT_SHAPE_CYLENDRICAL.
         */
        double contact_h_min, contact_h_max;

	/* Rectangular contact shape in meters for contact shape
	 * SAR_CONTACT_SHAPE_RECTANGULAR (left hand rule).
	 */
	double	contact_x_min, contact_x_max,
		contact_y_min, contact_y_max,
		contact_z_min, contact_z_max;

	/* Value records for the object's inversed heading for the trig
	 * functions, this is used to speed up rotations about heading
	 * for SAR_CONTACT_SHAPE_RECTANGULAR shape contact.
	 *
	 * Ie cos_heading = cos(-heading) and sin_heading = sin(-heading)
	 */
	double cos_heading, sin_heading;

} sar_contact_bounds_struct;


/*
 *	Rotor/propellar structure:
 */
typedef struct {

/* Can pitch forward for duel propulsion direction aircraft like the
 * Ospray.
 */
#define SAR_ROTOR_FLAG_CAN_PITCH	(1 << 1)

/* Is in forward prop motion position (pitched forwards if true).
 * Note that this has no affect on actual flight physics, the member
 * flight_model_type on the aircraft structure will determine the
 * correct flight model type.
 */
#define SAR_ROTOR_FLAG_PITCH_STATE	(1 << 2)

/* Does not rotate (spin) about its Z axis, used for stationary
 * rotor housing.
 */
#define SAR_ROTOR_FLAG_NO_ROTATE	(1 << 3)


	sar_obj_flags_t flags;

	/* Offset relative to center of object. */
        sar_position_struct pos;

	/* Rotation (always spins on Y axis). */
        sar_direction_struct dir;

	/* Radius in meters, for knowing size to make rotor wash. */
	double radius;

	/* Spin position, 0 to (sar_grad_anim_t)-1. */
	sar_grad_anim_t	anim_pos;

        sar_visual_model_struct *visual_model;

	/* Rotor pitch rotate position, 0 to (sar_grad_anim_t)-1,
	 * where (sar_grad_anim_t)-1 is pitched forwards
	 * (when SAR_ROTOR_FLAG_PITCH_STATE is set).
	 */
	sar_grad_anim_t pitch_anim_pos;

	/* Rotor wash animation position. */
	sar_grad_anim_t rotor_wash_anim_pos;

	/* Rotor wash texture number on scene, can be -1. */
	int rotor_wash_tex_num;

} sar_obj_rotor_struct;


/*
 *	Landing gear structure:
 */
typedef struct {

#define SAR_LGEAR_FLAG_STATE	(1 << 0)	/* True if gear is up. */
#define SAR_LGEAR_FLAG_FIXED	(1 << 1)	/* True if fixed (always down). */
#define SAR_LGEAR_FLAG_DAMAGED	(1 << 2)	/* True if damaged. */
#define SAR_LGEAR_FLAG_MISSING	(1 << 3)	/* Non-existant. */
#define SAR_LGEAR_FLAG_SKI	(1 << 4)

	sar_obj_flags_t flags;

	/* Offset relative to center of object. */
	sar_position_struct pos;

	/* Gear extended direction (retracted direction is 0 0 0,
	 * values are in radians and can be negative to indicate
	 * rotating to in the opposite direction.
	 */
	sar_direction_struct extended_dir;

	sar_grad_anim_t	anim_pos,	/* 0 to (sar_grad_anim_t)-1. */
			anim_rate;	/* Units per cycle. */

	sar_visual_model_struct *visual_model;

} sar_obj_landing_gear_struct;


/*
 *	Door structure:
 */
typedef struct {

#define SAR_DOOR_FLAG_STATE	(1 << 0)        /* Door opened if set. */
#define SAR_DOOR_FLAG_FIXED	(1 << 1)	/* True if fixed (always closed). */
#define SAR_DOOR_FLAG_LOCKED	(1 << 2)	/* True if locked. */

        sar_obj_flags_t flags;

	/* Positions, offsetted from object center. */
        sar_position_struct pos_closed, pos_opened;

	sar_direction_struct dir_opened;

	sar_grad_anim_t	anim_pos,
			anim_rate;

	sar_visual_model_struct *visual_model;

} sar_obj_door_struct;


/*
 *	Rescue hoist (role and basket) structure:
 */
typedef struct {

	sar_obj_flags_t flags;

	/* Offset from center of object. */
	sar_position_struct offset;

	/* Position (not offset) of basket. */
	sar_position_struct pos;

	/* Direction of basket. */
	sar_direction_struct dir;

	/* Rope extention (if <= 0 then implies it's `in') in meters. */
	double rope_cur, rope_max;
	double rope_rate;	/* In meters per cycle. */

	/* Cylendrical contact area of rescue basket, in meters. */
	double contact_radius;
	double contact_z_min, contact_z_max;

	/* Load capacity, in kg. */
	double capacity;

	/* Referance to occupant (people) object index numbers. */
	int *occupant;
	int total_occupants;

	/* Texture referance numbers on scene structure. */
	int side_tex_num, end_tex_num, bottom_tex_num;

} sar_obj_hoist_struct;


/*
 *	Object type codes:
 */
#define SAR_OBJ_TYPE_GARBAGE		0
#define SAR_OBJ_TYPE_STATIC		1
#define SAR_OBJ_TYPE_AUTOMOBILE		2
#define SAR_OBJ_TYPE_WATERCRAFT		3
#define SAR_OBJ_TYPE_AIRCRAFT		4

/* Cylendrical ground base with (optional) touchable heightfield. */
#define SAR_OBJ_TYPE_GROUND		6
/* Runway. */
#define SAR_OBJ_TYPE_RUNWAY		7
/* Helipad. */
#define SAR_OBJ_TYPE_HELIPAD		8

/* Human. */
#define SAR_OBJ_TYPE_HUMAN		9

/* Smoke puff. */
#define SAR_OBJ_TYPE_SMOKE		10
/* Fire. */
#define SAR_OBJ_TYPE_FIRE		11
/* Explosion. */
#define SAR_OBJ_TYPE_EXPLOSION		12

/* Premodeled object. */
#define SAR_OBJ_TYPE_PREMODELED		20


/*
 *	Flight model codes, used in member flight_model_type of
 *	aircraft structure sar_object_aircraft_struct.
 */
#define SAR_FLIGHT_MODEL_HELICOPTER	0
#define SAR_FLIGHT_MODEL_AIRPLANE	1
#define SAR_FLIGHT_MODEL_SLEW		2


/*
 *	SAR aircraft (helicopter/airplane) object sub structure:
 *
 *	This structure (and all object substructures) can be pointed to
 *	from member `data' in the sar_object_struct, determined by member
 *	`type'.
 */
typedef struct {

	/* Flight model type, one of SAR_FLIGHT_MODEL_*. */
	int flight_model_type;

	/* Previously set flight model type (for switching back and forth
	 * between flight model types.
	 */
	int last_flight_model_type;

	/* Sar flight model flight dynamics model structure. */
	SFMModelStruct *fdm;

        /* Current speed, in meters per cycle. */
        double speed;

	/* Stall speed, the speed at which a controllable stall begins.
	 * Loss in lift can still occure twice beyond this value, in
	 * meters per cycle.
	 */
	double speed_stall, stall_coeff;

	/* Maximum speed (this is *several* times the rated max), in
	 * meters per cycle.
	 */
	double speed_max;

	/* Current velocity vector representing velocity in object
	 * relative vector direction, in meters per cycle.
	 */
        sar_position_struct vel;

	/* Minimum drag in meters per cycle, must be non-negative. */
	double min_drag;

	/* Acceleration response for when flight_model_type is set to
	 * SAR_FLIGHT_MODEL_HELICOPTER. Higher values produce less
	 * responsiveness).
	 */
	sar_position_struct accel_responsiveness;

	/* Acceleration response for when flight_model_type is set to
	 * SAR_FLIGHT_MODEL_AIRPLANE. Higher values produce less
	 * responsiveness).
	 */
	sar_position_struct airplane_accel_responsiveness;


	/* Attitude change rates, in radians per cycle. */
	sar_direction_struct attitude_change_rate;

	/* Pitch and bank leveling, in radians per cycle. */
	double pitch_leveling, bank_leveling;


        /* Cockpit offset position in object relative vector
	 * direction, in meters.
	 */
        sar_position_struct cockpit_offset_pos;

	/* Cockpit visual model. */
	sar_visual_model_struct *visual_model_cockpit;

	/* Belly to object relative xy plane in meters. */
        double belly_height;

	/* Height of landing gear in meters. */
        double gear_height;

	/* Mass. */
	double dry_mass;		/* In kg. */
	double fuel_rate;		/* In kg per cycle. */
	double fuel, fuel_max;		/* In kg. */
	int crew, passengers, passengers_max;

	/* Engine data. */
	int engine_state;		/* One of SAR_ENGINE_STATE_*. */
	time_t next_engine_on;		/* Time till engine starts up. */
        double throttle;                /* 0.0 to 1.0. */
        double engine_power;            /* In kg * m / cycle^2. */
	int engine_can_pitch;		/* 1 if engines can pitch forwards/up. */
	char *engine_inside_sndpath;
	void *engine_inside_sndplay;
	char *engine_outside_sndpath;
	void *engine_outside_sndplay;

	/* Service ceiling, in meters. */
	double service_ceiling;

	/* Shadow visual model. */
	sar_visual_model_struct *visual_model_shadow;

	/* Distance from object's center to touchable ground in
	 * meters (usually negative).
	 */
	double center_to_ground_height;


	/* Rotors. */
	sar_obj_rotor_struct **rotor;
	int total_rotors;
	int rotor_wash_tex_num;	/* Texture number on scene struct. */

	/* Landing gears. */
	sar_obj_landing_gear_struct **landing_gear;
	int total_landing_gears;
	int landed;	/* Indicates object is on ground if true. */
	double gturn_radius;	/* Distance from farthest non-turning gear to
				 * turning gear in meters, can be negative.
				 * If 0 then it implies it cannot turn.
				 */
	/* Ground turning velocity optimul and maximum in meters per
	 * cycle.
	 */
	double gturn_vel_opt, gturn_vel_max;

	/* Rescue door. */
	sar_obj_door_struct rescue_door;

	/* Air brakes and wheel brakes:
	 * -1   non-existant
	 * 0    off
	 * 1    on
	 * 2    locked (parking brakes on)
	 */
	int	air_brakes_state,
		wheel_brakes_state;

	/* Air worthy state, one of SAR_AIR_WORTHY_*. */
	int air_worthy_state;

	/* Rescue hoist. */
	sar_obj_hoist_struct hoist;

	/* Intercept way points. */
	sar_obj_intercept_struct **intercept;
	int total_intercepts, cur_intercept;

} sar_object_aircraft_struct;

/*
 *	SAR ground elevation object sub structure:
 *
 *      This structure can be pointed to from member `data' in the
 *      sar_object_struct, determined by member `type'.
 */
typedef struct {

	/* Elevation from (above) local MSL in meters. If MSL was
	 * set to 10 meters and this elevation was set to 3 meters then
	 * the resulting elevation would be 13 meters to any object within
	 * contact of the ground object.
	 *
	 * Note that the ground object itself completely ignores the
	 * scene structure defined ground_elevation_msl value.
	 */
	double elevation;

	/* Translation of heightfield from center of object. */
	double x_trans, y_trans, z_trans;

	/* Heightfield data (only used if z_point_value is not NULL). */
	double	x_len,		/* grid_points_x * grid_x_spacing */
		y_len;		/* grid_points_y * grid_y_spacing */
	int	grid_points_x,	/* Number of grid points. */
		grid_points_y,
		grid_points_total;	/* grid_points_x * grid_points_y */
	double	grid_x_spacing,	/* Size of each grid in meters. */
		grid_y_spacing,
		grid_z_spacing;
	double	*z_point_value;	/* Heightfield z height map, each point
				 * value came from (image_data_pixel) /
				 * 0xff * grid_z_spacing.
				 */

        /* Value records for this ground object's inversed heading for the
	 * trig functions, this is used to speed up rotations of heading
         * for checking an object over this ground object's heightfield
	 * surface. It also limits ground objects to only have its
	 * heading rotated.
         *
         * Ie cos_heading = cos(-heading) and sin_heading = sin(-heading)
         */
        double cos_heading, sin_heading;

} sar_object_ground_struct;

/*
 *	SAR smoke puff sub structure:
 */
typedef struct {

	/* Displayed size of explosion in meters at time of birth. */
	double radius;

        /* Texture number on scene structure. */
        int tex_num;

} sar_object_smoke_struct;

/*
 *	SAR explosion sub structure:
 */
typedef struct {

	/* Explosions do not have model data. */

	/* Displayed size of explosion in meters. */
	double radius;

	int cur_frame;		/* Current frame. */
	time_t next_frame_inc;	/* Time to increment next frame. */

	/* Texture number on scene structure. */
	int tex_num;

	/* Referance object that this explosion is to follow
	 * (can be -1 for none, do not follow).
	 */
	int ref_object;

} sar_object_explosion_struct;


/*
 *      SAR runway sub structure:
 */
typedef struct {

#define SAR_RUNWAY_FLAG_NORTH_LABEL	(1 << 1)
#define SAR_RUNWAY_FLAG_SOUTH_LABEL	(1 << 2)
#define SAR_RUNWAY_FLAG_NORTH_DISPTHRES	(1 << 3)
#define SAR_RUNWAY_FLAG_SOUTH_DISPTHRES	(1 << 4)
#define SAR_RUNWAY_FLAG_EDGE_LIGHTING	(1 << 5)
#define SAR_RUNWAY_FLAG_NORTH_LIGHTING	(1 << 6)
#define SAR_RUNWAY_FLAG_SOUTH_LIGHTING	(1 << 7)

	sar_obj_flags_t flags;

	/* Size in meters. */
	double length, width;

	/* Runway end labels. */
	char *north_label, *south_label;

	/* Dash spacing in meters. */
	double	dash_on_spacing,
		dash_off_spacing;

	/* Lighting edge spacing in meters. */
	double light_spacing;

	/* Threshold displacements, from end to threshold in meters. */
	double	north_threshold_displacement,
		south_threshold_displacement;

        /* Texture number on scene structure. */
        int tex_num;

} sar_object_runway_struct;


/*
 *	SAR helipad sub structure:
 */
typedef struct {

#define SAR_HELIPAD_FLAG_LABEL		(1 << 1)
#define SAR_HELIPAD_FLAG_EDGE_LIGHTING	(1 << 2)

        sar_obj_flags_t flags;

        /* Size in meters. */
        double length, width;

        /* Labels. */  
        char *label;

        /* Lighting edge spacing in meters. */
        double light_spacing;

        /* Texture number on scene structure. */
        int tex_num;

} sar_object_helipad_struct;

/*
 *	SAR human structure:
 */
typedef struct {

#define SAR_HUMAN_FLAG_NEED_RESCUE	(1 << 1)
#define SAR_HUMAN_FLAG_SIT		(1 << 2)	/* Base at tush. */
#define SAR_HUMAN_FLAG_SIT_DOWN		(1 << 3)	/* Base at tush, feet out. */
#define SAR_HUMAN_FLAG_SIT_UP		(1 << 4)	/* Base at feet. */
#define SAR_HUMAN_FLAG_LYING		(1 << 5)
#define SAR_HUMAN_FLAG_ALERT		(1 << 6)	/* Is awake. */
#define SAR_HUMAN_FLAG_AWARE		(1 << 7)	/* Knows of surroundings. */
#define SAR_HUMAN_FLAG_IN_WATER		(1 << 8)
#define SAR_HUMAN_FLAG_RUN_TOWARDS	(1 << 9)
#define SAR_HUMAN_FLAG_RUN_AWAY		(1 << 10)
#define SAR_HUMAN_FLAG_GRIPPED		(1 << 11)	/* In rescue basket for example. */

	sar_obj_flags_t flags;

	double weight;	/* In kg. */

	sar_grad_anim_t anim_pos, anim_rate;

	/* Body colors. */
	sar_color_struct	face_color,
				hair_color,
				torso_color,
				hips_color,
				legs_color,
				feet_color,
				arms_color,
				hands_color;

	/* Water ripples texture number on scene structure. */
	int water_ripple_tex_num;

	/* Referance to object running towards or away from.
	 * Can be -1, and only works if either (pick only one)
	 * flags SAR_HUMAN_FLAG_RUN_TOWARDS or SAR_HUMAN_FLAG_RUN_AWAY
	 * is set.
	 */
	int intercepting_object;

	/* Messages. */
	char *mesg_hoist_in;	/* When hoisted into the aircraft. */

} sar_object_human_struct;

/*
 *	SAR premodeled object:
 */
typedef struct {

#define SAR_OBJ_PREMODELED_BUILDING		1
#define SAR_OBJ_PREMODELED_CONTROL_TOWER	2
#define SAR_OBJ_PREMODELED_HANGAR		3

	int type;	/* One of SAR_OBJ_PREMODELED_*. */

	/* Values, depending on type, not all may be applicateable. */
	double width, length, height;

	/* Animation. */
	sar_grad_anim_t anim_pos, anim_rate;

        /* Texture number on scene structure. */
#define SAR_OBJ_PREMODEL_MAX_TEXTURES	10
        int tex_num[SAR_OBJ_PREMODEL_MAX_TEXTURES];

} sar_object_premodeled_struct;

/*
 *	SAR core object structure:
 */
typedef struct {

	int type;	/* Must be first member, one of SAR_OBJ_TYPE_*. */

#define SAR_OBJ_FLAG_NO_DEPTH_TEST	(1 << 1)
#define SAR_OBJ_FLAG_HIDE_DAY_MODEL	(1 << 2)	/* Hide day model when not day. */
#define SAR_OBJ_FLAG_HIDE_DAWN_MODEL	(1 << 3)	/* Hide dawn model when not dawn. */
#define SAR_OBJ_FLAG_HIDE_DUSK_MODEL	(1 << 4)	/* Hide dusk model when not dusk. */
#define SAR_OBJ_FLAG_HIDE_NIGHT_MODEL	(1 << 5)	/* Hide night model when not night. */

	sar_obj_flags_t flags;

	/* Name of object (can be NULL). */
	char *name;

        /* Current position and direction. */
        sar_position_struct pos;
        sar_direction_struct dir;

	/* Radius of how far this object can be seen, in meters. */
	double range;

	/* Distance from sea level to touchable ground under
	 * object, in meters. (Not applicateable object type
	 * SAR_OBJ_TYPE_GROUND).
	 */
	double ground_elevation_msl;


	/* Contact and crash bounds. */
	sar_contact_bounds_struct *contact_bounds;


	/* Time stamp of when this object was created in milliseconds
	 * and in seconds (respectivly). The value in milliseconds may
	 * not be accurate due to the Cinderella issue.
	 */
	time_t	birth_time_ms,
		birth_time_sec;

	/* When this object `dies'  (0 if lives forever). */
	time_t life_span;

	/* Hit points. */
	double hit_points, hit_points_max;

	/* Visual models for day, dusk, night. */
	sar_visual_model_struct *visual_model;		/* Day. */
	sar_visual_model_struct *visual_model_dawn;	/* Dawn. */
	sar_visual_model_struct *visual_model_dusk;	/* Dusk. */
	sar_visual_model_struct *visual_model_night;	/* Night. */

	/* Lights. */
	sar_light_struct **light;
	int total_lights;

	/* Pointer to additional data specific to the object type,
	 * the structure's type is specified by this structure's member
	 * type.
	 */
	void *data;

} sar_object_struct;



/*
 *	Scene ground base structure, contains visual models and tiling
 *	values for displaying of the ground base.
 */
typedef struct {

	/* Note, ground base moves with camera in modulous increments
	 * of the tiled width and height.
	 */

	/* Tiling size in meters. */
        int tile_width, tile_height;

	/* Tiling limited to this range in meters from origin of
	 * tiling.
	 */
	double close_range;


	/* Simple (farther away) solid color base plane and visual
	 * model.
	 */
	sar_color_struct color;
        sar_visual_model_struct *visual_model_simple;

	/* Close up texture tiled base plane visual model. */
	sar_visual_model_struct *visual_model_close;

} sar_scene_base_struct;


/*
 *	Scene horizon structure:
 */
typedef struct {

	/* Records the last time of day since midnight from
	 * the scene structure in 30 minute units as a whole number for
	 * horizon texture regeneration.
	 */
	int last_tod;

	/* Gradient textures, 0 is most highlighted and
	 * total_textures - 1 is darkest.
	 */
	v3d_texture_ref_struct **texture;
	int total_textures;

} sar_scene_horizon_struct;


/*
 *	Scene cloud layer structure:
 */
typedef struct {

        /* Note, cloud layer xy moves with camera in modulous increments
         * of the tiled width and height.
         */

        /* Tiling size in meters. */
	int tile_width, tile_height;

        /* Tiling limited to this range in meters from origin of
         * tiling.
         */
        double close_range;

	char *tex_name;
	sar_visual_model_struct *visual_model;

	/* Height from MSL in meters. */
	double z;

} sar_cloud_layer_struct;


/*
 *	SAR scene structure:
 */
typedef struct {

	/* Time of day (in seconds) since midnight. */
	double tod;

	/* Time of day code, one of SARSceneTODCode*. */
#define SARSceneTODCodeDay		1
#define SARSceneTODCodeDawn		2
#define SARSceneTODCodeNight		3
#define SARSceneTODCodeDusk		4
	int tod_code;

        /* Base direction, all objects in scene get translated by
         * this.
         */
        sar_direction_struct base_dir;   

	/* MSL elevation in meters. */
	double msl_elevation;

	/* Scene on planet values. */
	sar_dms_t	dms_x_offset,	/* -140.0 to 140.0. */
			dms_y_offset;	/* -90.0 to 90.0. */
	double planet_radius;		/* In meters. */

	/* List of visual models (structures containing GL list
	 * referances) each object or resource that has a visual
	 * model (has a GL list basically) will point to one of
	 * these structures. Only these structures may be deallocated.
	 */
	sar_visual_model_struct **visual_model;
	int total_visual_models;

	/* Sky and horizon gradient colors. */
	sar_color_struct	sky_nominal_color,
				sky_brighten_color,
				sky_darken_color;

	sar_color_struct	star_color,
				moon_color;

	/* Atmosphere (when enabled). */
	double	atmosphere_dist_coeff,	/* Coeff muliplied by max visiblity, */
		atmosphere_density_coeff;

	/* Referance to player object. */
	int player_obj_num;
	sar_object_struct *player_obj_ptr;

	/* Pointers to SAR_OBJ_TYPE_GROUND objects, you may free
	 * the pointer array but not the pointer to structures!!
	 */
	sar_object_struct **ground_object;
	int total_ground_objects;

	/* Pointers to SAR_OBJ_TYPE_HUMAN objects that *need rescue*.
	 * You may free the pointer array but not the pointers to structures!!
	 */
	sar_object_struct **human_need_rescue_object;
        int total_human_need_rescue_objects;


	/* Camera referance. */
#define SAR_CAMERA_REF_COCKPIT	0
#define SAR_CAMERA_REF_SPOT	1
#define SAR_CAMERA_REF_TOWER	2
#define SAR_CAMERA_REF_MAP	3
#define SAR_CAMERA_REF_HOIST	4	/* Rescue hoist view. */

	int camera_ref;

	/* Camera field of view, in radians along Z axis (or
	 * you can say `about the X axis').
	 */
	double camera_fovz;

	/* Direction camera is looking while set to
	 * SAR_CAMERA_REF_COCKPIT. The bank member is ignored.
	 */
	sar_direction_struct camera_cockpit_dir;

	/* Direction from object to camera while set to
	 * SAR_CAMERA_REF_SPOT. The bank member is ignored.
	 */
	sar_direction_struct camera_spot_dir;
	double camera_spot_dist;

	/* Distance for SAR_CAMERA_REF_HOIST. */
	sar_direction_struct camera_hoist_dir;
	double camera_hoist_dist;

	/* Position of camera (may not be exact). */
	sar_position_struct camera_pos;

	/* Target object that camera is tracking (can be -1). */
	int camera_target;


	/* Primary light position. */
	sar_position_struct light_pos;
	sar_color_struct light_color;


	/* Base ground. */
	sar_scene_base_struct base;

	/* Horizon. */
	sar_scene_horizon_struct horizon;

	/* Cloud layers. */
	sar_cloud_layer_struct **cloud_layer;
	int total_cloud_layers;

	/* Loaded textures list. */
	v3d_texture_ref_struct **texture_ref;
	int total_texture_refs;

	/* Default sounds object paths. */
	char	*land_wheel_skid_sndpath,
		*land_ski_skid_sndpath,		/* Skis skidding. */
		*land_ski_sndpath,		/* Skis just touching down. */
		*land_belly_sndpath,
		*crash_obstruction_sndpath,
		*crash_ground_sndpath;

	/* Messages. */
	char **message;
	int total_messages;
	time_t message_display_untill;	/* In milliseconds. */

	/* Name of object and/or description of what the camera is
	 * showing.
	 */
	char *camera_ref_title;
	time_t camera_ref_title_display_untill;

	/* Flight dynamics model realm structure. */
	SFMRealmStruct *realm;

} sar_scene_struct;



/*
 *	Code segment to get typed pointer to object's extended
 *	data.
 */
/*
	sar_object_struct *obj_ptr;
        sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	sar_object_ground_struct *obj_ground_ptr = NULL;
	sar_object_runway_struct *obj_runway_ptr = NULL;
        sar_object_helipad_struct *obj_helipad_ptr = NULL;
	sar_object_smoke_struct *obj_smoke_ptr = NULL;
	sar_object_explosion_struct *obj_explosion_ptr = NULL;
	sar_object_human_struct *obj_human_ptr = NULL;
 */
#define SAR_GET_EXT_DATA_PTR	\
/* obj_ptr must not be NULL. */ \
switch(obj_ptr->type) \
{ \
 case SAR_OBJ_TYPE_AIRCRAFT: \
  obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data; \
  break; \
 case SAR_OBJ_TYPE_GROUND: \
  obj_ground_ptr = (sar_object_ground_struct *)obj_ptr->data; \
  break; \
 case SAR_OBJ_TYPE_RUNWAY: \
  obj_runway_ptr = (sar_object_runway_struct *)obj_ptr->data; \
  break; \
 case SAR_OBJ_TYPE_HELIPAD: \
  obj_helipad_ptr = (sar_object_helipad_struct *)obj_ptr->data; \
  break; \
 case SAR_OBJ_TYPE_SMOKE: \
  obj_smoke_ptr = (sar_object_smoke_struct *)obj_ptr->data; \
  break; \
 case SAR_OBJ_TYPE_EXPLOSION: \
  obj_explosion_ptr = (sar_object_explosion_struct *)obj_ptr->data; \
  break; \
 case SAR_OBJ_TYPE_HUMAN: \
  obj_human_ptr = (sar_object_human_struct *)obj_ptr->data; \
  break; \
} \



/* In objutils.c */
extern void SARSetDirectionRadians(
	sar_direction_struct *dir,
	double heading, double pitch, double bank
);

extern int SARObjIsAllocated(
        sar_object_struct **ptr, int total,
        int n
);
extern int SARGetObjectNumberFromPointer(
        sar_scene_struct *scene,
        sar_object_struct **ptr, int total,
        sar_object_struct *obj_ptr
);
extern sar_object_struct *SARObjMatchPointerByName(
        sar_scene_struct *scene,
        sar_object_struct **ptr, int total,
        const char *name, int *obj_num
);

extern int SARIsTextureAllocated(
        sar_scene_struct *scene,
        int n
);
extern v3d_texture_ref_struct *SARGetTextureRefByName(
	sar_scene_struct *scene, const char *name
);
extern int SARGetTextureRefNumberByName(
        sar_scene_struct *scene, const char *name
);

extern sar_visual_model_struct *SARVisualModelNew(
        sar_scene_struct *scene,
        const char *filename, const char *name
);
extern void *SARVisualModelNewList(sar_visual_model_struct *vmodel);
extern int SARVisualModelGetRefCount(sar_visual_model_struct *vmodel);
extern void SARVisualModelRef(sar_visual_model_struct *vmodel);
extern void SARVisualModelUnref(sar_visual_model_struct *vmodel);
extern void SARVisualModelCallList(sar_visual_model_struct *vmodel);
extern void SARVisualModelDeleteAll(sar_scene_struct *scene);

extern sar_cloud_layer_struct *SARCloudLayerCreate(
	sar_scene_struct *scene,
	int tile_width, int tile_height,
	double close_range,
	double altitude,
	const char *tex_name
);          
extern void SARCloudLayerDestroy(
        sar_cloud_layer_struct *cloud_layer_ptr
);
extern int SARObjAddToGroundList(
        sar_scene_struct *scene,
        sar_object_struct *obj_ptr
);
extern void SARObjRemoveFromGroundList(
        sar_scene_struct *scene,
        sar_object_struct *obj_ptr
);
extern int SARObjAddToHumanRescueList(   
        sar_scene_struct *scene,
        sar_object_struct *obj_ptr
);
extern void SARObjRemoveFromHumanRescueList(
        sar_scene_struct *scene,
        sar_object_struct *obj_ptr
);

extern int SARObjAddContactBoundsSpherical(
        sar_object_struct *obj_ptr,
        sar_obj_flags_t crash_flags, int crash_type,
        double contact_radius
);
extern int SARObjAddContactBoundsCylendrical(
        sar_object_struct *obj_ptr,
        sar_obj_flags_t crash_flags, int crash_type,
        double contact_radius,
        double contact_h_min, double contact_h_max
);
extern int SARObjAddContactBoundsRectangular(
        sar_object_struct *obj_ptr,
        sar_obj_flags_t crash_flags, int crash_type,
        double contact_x_min, double contact_x_max,
        double contact_y_min, double contact_y_max,
        double contact_z_min, double contact_z_max
);

extern int SARObjCreateIntercept(
        sar_scene_struct *scene,
        sar_obj_intercept_struct ***ptr, int *total,
        sar_obj_flags_t flags,
        double x, double y, double z,
        double radius,
        double urgency,
        const char *name
);
extern int SARObjCreateLandingGear(
        sar_scene_struct *scene,
        sar_obj_landing_gear_struct ***ptr, int *total
);
extern int SARObjCreateRotor(  
        sar_scene_struct *scene,
        sar_obj_rotor_struct ***ptr, int *total
);
extern int SARObjCreate(
	sar_scene_struct *scene,
        sar_object_struct ***ptr, int *total,
        int type
);

extern void SARObjDeleteIntercepts(
	sar_scene_struct *scene,
        sar_obj_intercept_struct ***ptr, int *total
);
extern void SARObjDeleteLights(
	sar_scene_struct *scene,
	sar_light_struct ***ptr, int *total
);
extern void SARObjDeleteLandingGears(
        sar_scene_struct *scene,
        sar_obj_landing_gear_struct ***ptr, int *total
);
extern void SARObjDeleteRotors(
        sar_scene_struct *scene,
        sar_obj_rotor_struct ***ptr, int *total
);
extern void SARObjDeleteDoors(
        sar_scene_struct *scene,
        sar_obj_door_struct ***ptr, int *total
);
extern void SARObjDelete(
	void *core_ptr,
        sar_object_struct ***ptr, int *total,
        int n
);

extern void SARObjGenerateTilePlane(
        double min, double max, /* In meters. */
        double tile_width, double tile_height
);


#endif	/* OBJ_H */
