#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <math.h>

#include "../include/os.h"
#ifdef __MSW__
# include <windows.h>
#endif

#include <GL/gl.h>

#include "../include/fio.h"
#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/disk.h"

#include "sfm.h"

#include "sarreality.h"
#include "obj.h"
#include "simmanage.h"
#include "simcb.h"
#include "simutils.h"
#include "weather.h"

#include "sar.h"


void SARSceneDestroy(
	sar_core_struct *core_ptr,
	sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total
);
void SARSceneLoadLocationsToList(
        sar_core_struct *core_ptr,
        sar_menu_struct *menu_ptr,
        sar_menu_list_struct *list_ptr, int list_num,
        const char *filename
);
int SARSceneAddPlayerObject(
	sar_core_struct *core_ptr,
	sar_scene_struct *scene,
	const char *model_file,
	sar_position_struct *pos,
	sar_direction_struct *dir
);
int SARSceneLoadFromFile(
	sar_core_struct *core_ptr,
	sar_scene_struct *scene,
	const char *filename,
	const char *weather_preset_name,
	void *client_data,
	int (*progress_func)(void *, long, long)
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))

#define ISCOMMENT(c)	((c) == SAR_COMMENT_CHAR)
#define ISCR(c)		(((c) == '\n') || ((c) == '\r'))


/*
 *	Quick converter from degrees to radians without sanitization:
 */
#define DEGTORAD(d)	((d) * PI / 180)


/*
 *	Deallocates all resources on the scene structure and all
 *	objects on the given objects pointer array.
 *
 *	Does not deallocate the scene structure itself, but the
 *	scene structure will be reset.
 */
void SARSceneDestroy(
	sar_core_struct *core_ptr, sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total
)
{
	int i;
	sar_scene_horizon_struct *horizon_ptr;


	if(core_ptr == NULL)
	    return;


	/* Check if total objects is not NULL (which implies ptr is not
	 * NULL).
	 */
	if(total != NULL)
	{
	    /* Deallocate objects pointer array from last to first. */
	    for(i = (*total) - 1; i >= 0; i--)
	        SARObjDelete(core_ptr, ptr, total, i);

	    (*total) = 0;
	}
	/* Be sure to free objects pointer array. */
	if(ptr != NULL)
	{
	    free(*ptr);
	    (*ptr) = NULL;
	}


	/* Is scene structure valid? */
	if(scene != NULL)
	{
	    /* Deallocate all resources on scene structure. */

	    /* Ground base visual model simple (far away). */
	    SARVisualModelUnref(scene->base.visual_model_simple);
	    scene->base.visual_model_simple = NULL;

	    /* Ground base visual model complext (close up). */
            SARVisualModelUnref(scene->base.visual_model_close);
            scene->base.visual_model_close = NULL;


            /* Cloud layers. */
	    for(i = 0; i < scene->total_cloud_layers; i++)
		SARCloudLayerDestroy(scene->cloud_layer[i]);
	    free(scene->cloud_layer);
	    scene->cloud_layer = NULL;
	    scene->total_cloud_layers = 0;

	    /* Horizon. */
	    horizon_ptr = &scene->horizon;
	    for(i = 0; i < horizon_ptr->total_textures; i++)
		V3DTextureDestroy(horizon_ptr->texture[i]);

	    free(horizon_ptr->texture);
	    horizon_ptr->texture = NULL;

	    horizon_ptr->total_textures = 0;


	    /* Ground objects list. */
	    free(scene->ground_object);

            /* Human that need rescue objects list. */
            free(scene->human_need_rescue_object);

	    /* Visual models, all visual models should have been
	     * unref'ed by now. So here we actually deallocate them
	     * and destroy the GL lists.
	     */
	    SARVisualModelDeleteAll(scene);

	    /* Delete all textures (deallocated by V3DTextureDestroy()). */
	    for(i = 0; i < scene->total_texture_refs; i++)
		V3DTextureDestroy(scene->texture_ref[i]);
	    free(scene->texture_ref);	/* Free pointer array. */

	    /* Default sound object paths. */
	    free(scene->land_wheel_skid_sndpath);
            free(scene->land_ski_skid_sndpath);
            free(scene->land_ski_sndpath);
	    free(scene->land_belly_sndpath);
            free(scene->crash_obstruction_sndpath);
            free(scene->crash_ground_sndpath);

	    /* Messages. */
	    StringFreeArray(scene->message, scene->total_messages);

	    /* Camera referance name/description title string. */
	    free(scene->camera_ref_title);

	    /* Realm structure. */
	    if(scene->realm != NULL)
		SFMShutdown(scene->realm);

	    /* Reset the rest of the scene structure. */
	    memset(scene, 0x00, sizeof(sar_scene_struct));
	}

	return;
}

/*
 *      Loads the list of registered locations specified in the scene
 *      file and stores them in the list object. The contents of the
 *      list object will be cleared first.
 */
void SARSceneLoadLocationsToList(
        sar_core_struct *core_ptr,
        sar_menu_struct *menu_ptr,
        sar_menu_list_struct *list_ptr, int list_num,
        const char *filename
)
{
        int i;
        FILE *fp;
        char *strptr;


        if(list_ptr == NULL)
            return;

        /* Delete current items if any on the list object, first
	 * delete their client_data structures.
	 */
        for(i = 0; i < list_ptr->total_items; i++)
            SARDeleteListItemData(list_ptr->item[i]);
        SARMenuListDeleteAllItems(menu_ptr, list_num);


        if(filename == NULL)
            return;

	/* Open scene file. */
        fp = FOpen(filename, "rb");
        if(fp == NULL)
            return;

        /* Seek to each register_location parameter. */
        while(!FSeekToParm(fp, "register_location", SAR_COMMENT_CHAR, '\0'))
        {
            strptr = FGetString(fp);
            if(strptr != NULL)
            {
                int strc, n;
                char **strv;
                const char *loc_name = "Unknown";
                sar_menu_list_item_data_struct *li_data_ptr;


		/* Explode arguments. */
                strv = strexp(strptr, &strc);

		/* Allocate a new list item data structure. */
                li_data_ptr = (sar_menu_list_item_data_struct *)calloc(
                    1,
                    sizeof(sar_menu_list_item_data_struct)
                );
                if(li_data_ptr != NULL)
                {
		    /* Set and seek loc_name past numbers and spaces
		     * so that it points to the name.
		     */
                    loc_name = (const char *)strptr;
                    while(isdigit(*loc_name) ||   
                          ISBLANK(*loc_name) ||
                          ((*loc_name) == '-') ||
                          ((*loc_name) == '.') ||
                          ((*loc_name) == ',')
                    )
                        loc_name++;

		    /* Get position of this location. */
                    if(strc > 2)   
                    { 
                        li_data_ptr->pos.x = atof(strv[0]);
                        li_data_ptr->pos.y = atof(strv[1]);
                        li_data_ptr->pos.z = SFMFeetToMeters(atof(strv[2]));
                    }

		    /* Get direction of this location. */
                    if(strc > 5)
                    {
                        li_data_ptr->dir.heading = SFMDegreesToRadians(
                            atof(strv[3])
                        );  
                        li_data_ptr->dir.pitch = SFMDegreesToRadians(
                            atof(strv[4])
                        );
                        li_data_ptr->dir.bank = SFMDegreesToRadians(
                            atof(strv[5])
                        );
                    } 

		    /* Set new name of this location. */
                    li_data_ptr->name = StringCopyAlloc(loc_name);

		    /* Add this location to the list. */
                    n = SARMenuListAppendItem(
                        menu_ptr, list_num,
                        loc_name, li_data_ptr,
                        0
                    );
                    if(n < 0)
                    {
                        fprintf(stderr,
 "Error appending list item for starting location `%s'\n",
                            loc_name
                        );

                        free(li_data_ptr->name);
                        free(li_data_ptr);
                        li_data_ptr = NULL;
                    }
                }

                StringFreeArray(strv, strc);
                free(strptr);
		strptr = NULL;
            }
        }

	/* Close scene file. */
        FClose(fp);

        return;
}


/*
 *	Adds a player objec to the scene using the given model
 *	file and initial position and direction. Returns non-zero
 *	on error.
 *
 *	Updates the player object referances on the scene structure.
 *
 *	All inputs must be valid (except for pos and dir).
 */
int SARSceneAddPlayerObject(
	sar_core_struct *core_ptr,
	sar_scene_struct *scene,
        const char *model_file,
	sar_position_struct *pos,
	sar_direction_struct *dir
)
{
	int obj_num;
	sar_object_struct *obj_ptr;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_position_struct lpos;
	sar_direction_struct ldir;


	if((core_ptr == NULL) ||
           (scene == NULL) ||
	   (model_file == NULL)
	)
	    return(-1);

	/* Create player object on to scene. */
	obj_num = SARObjCreate(
	    scene, &core_ptr->object, &core_ptr->total_objects,
	    SAR_OBJ_TYPE_AIRCRAFT
	);
        obj_ptr = ((obj_num < 0) ? NULL : core_ptr->object[obj_num]);

	if(obj_ptr == NULL)
	    return(-1);

	/* Load its model file. */
        SARObjLoadFromFile(
            core_ptr, obj_num,
            model_file
	);

	/* Set object name to "player", so that it can be
	 * matched later by name (ie needed later in mission file parsing)
	 */
	free(obj_ptr->name);
	obj_ptr->name = strdup("player");

	/* Update player object referances on scene structure. */
	scene->player_obj_num = obj_num;
	scene->player_obj_ptr = obj_ptr;


	/* Set local position structure. */
	if(pos != NULL)
	    memcpy(&lpos, pos, sizeof(sar_position_struct));
	else
	    memset(&lpos, 0x00, sizeof(sar_position_struct));

        /* Move player up from ground level so that it does not added into
         * the scene while embedded in the ground (which may cause a
         * crash if the given pos is ontop of a building).
         */
	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
	    obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
	    if(obj_aircraft_ptr != NULL)
	    {
		lpos.z += MAX(obj_aircraft_ptr->belly_height, 0);
		lpos.z += MAX(obj_aircraft_ptr->gear_height, 0);
	    }
	    break;

/* Add other object types that need their position modified here. */

	}

	/* Set local direction structure. */
        if(dir != NULL)
            memcpy(&ldir, dir, sizeof(sar_direction_struct));
        else
            memset(&ldir, 0x00, sizeof(sar_direction_struct));


        /* Move player object to starting location and attitude. */
	SARSimWarpObject(scene, obj_ptr, &lpos, &ldir);

	return(0);
}

/*
 *	Loads a new scene from the specified file.
 *
 *	If the given buffers contain data then they will be deallocated
 *	first before the new data is loaded from file.
 *
 *	Initial GL states will be set to defaults here.
 */
int SARSceneLoadFromFile(
	sar_core_struct *core_ptr,
	sar_scene_struct *scene,
	const char *filename,
	const char *weather_preset_name,
	void *client_data,
	int (*progress_func)(void *, long, long)
)
{
	FILE *fp;
	char *strptr, *strptr2;
	int c, obj_num = -1, lines_read = 0, *total;
	long file_pos, file_size;
	gw_display_struct *display;
	sar_direction_struct *dir;
	sar_color_struct *color;
	sar_object_struct *obj_ptr = NULL, ***ptr;
	sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	sar_object_ground_struct *obj_ground_ptr = NULL;
	sar_scene_horizon_struct *horizon_ptr;
        double value[10];
	int argc;
	char **argv;
        struct stat stat_buf;


/* Macro to reset sar object substructure pointers to NULL. */
#define DO_RESET_SUBSTRUCTURE_PTRS      \
{ \
 obj_aircraft_ptr = NULL; \
 obj_ground_ptr = NULL; \
}


        if((core_ptr == NULL) ||
           (scene == NULL) ||
           (filename == NULL)
        )
            return(-1);

	display = core_ptr->display;

	ptr = &core_ptr->object;
	total = &core_ptr->total_objects;


	/* Check if file exists. */
        if(stat(filename, &stat_buf))
            return(-1);
	file_size = stat_buf.st_size;	/* Get file size. */

	/* Open file. */
        fp = FOpen(filename, "rb");
        if(fp == NULL)
            return(-1);


	/* Delete current scene and objects. */
	SARSceneDestroy(core_ptr, scene, ptr, total);


        /* Reset scene. */
	scene->tod = (12 * 3600);
	scene->tod_code = SARSceneTODCodeDay;
        scene->msl_elevation = 0.0;

	scene->dms_x_offset = 0;
	scene->dms_y_offset = 0;
	scene->planet_radius = SAR_DEF_PLANET_RADIUS;

	scene->visual_model = NULL;
	scene->total_visual_models = 0;

	color = &scene->sky_nominal_color;
	color->r = 1.0;
	color->g = 1.0;
	color->b = 1.0;
        color = &scene->sky_brighten_color;
        color->r = 1.0;
        color->g = 1.0;
        color->b = 1.0;
        color = &scene->sky_darken_color;
        color->r = 1.0;
        color->g = 1.0;
        color->b = 1.0;

        color = &scene->star_color;
        color->r = 1.0;
        color->g = 1.0;
        color->b = 1.0;
        color = &scene->moon_color;
        color->r = 1.0;
        color->g = 1.0;
        color->b = 1.0;

        scene->ground_object = NULL;
        scene->total_ground_objects = 0;
        
        memset(&scene->base_dir, 0x00, sizeof(sar_direction_struct));
        scene->player_obj_num = -1;
        scene->player_obj_ptr = NULL;

        scene->camera_ref = SAR_CAMERA_REF_COCKPIT;
        scene->camera_fovz = SFMDegreesToRadians(40);
	dir = &scene->camera_cockpit_dir;
	dir->heading = 0 * PI;
	dir->pitch = 0 * PI;
	dir->bank = 0 * PI;
	dir = &scene->camera_spot_dir;
	dir->heading = 1.25 * PI;
	dir->pitch = 1.92 * PI;
	dir->bank = 0 * PI;
        scene->camera_spot_dist = 20.0;
        dir = &scene->camera_hoist_dir;
	dir->heading = 0.75 * PI;
        dir->pitch = 1.92 * PI;
        dir->bank = 0 * PI;
	scene->camera_hoist_dist = 20.0;
        scene->camera_target = -1;

	/* Reset initial GL states. */
	if(display != NULL)
	{
	    StateGLResetAll(&display->state_gl);
	    StateGLEnable(&display->state_gl, GL_CULL_FACE);
	    StateGLFrontFace(&display->state_gl, GL_CCW);
	}
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);


	/* Allocate messages. */
	scene->total_messages = 4;
	scene->message = (char **)calloc(scene->total_messages, sizeof(char *));
	if(scene->message == NULL)
	    scene->total_messages = 0;

	scene->message_display_untill = 0;
	scene->camera_ref_title_display_untill = 0;

	/* Allocate a realm structure. */
	scene->realm = SFMInit(0, NULL);
	if(scene->realm == NULL)
	{
	    fprintf(stderr,
		"SARSceneLoadFromFile(): Could not create FDM realm.\n"
	    );
	}
	else
	{
	    SFMRealmStruct *realm = scene->realm;

	    /* Set up realm values. */
	    realm->lapsed_time = 0;
	    realm->time_compensation = 1.0;
	    realm->time_compression = 1.0;

	    realm->gravity = SFMDefaultGravity;

	    /* Set up realm callback data and callback function ptrs. */
	    realm->init_model_cb_client_data = core_ptr;
	    realm->init_model_cb = SARSimInitModelCB;
            realm->destroy_model_cb_client_data = core_ptr;
            realm->destroy_model_cb = SARSimDestroyModelCB;
            realm->airborne_cb_client_data = core_ptr;
            realm->airborne_cb = SARSimAirborneCB;
            realm->touch_down_cb_client_data = core_ptr;
            realm->touch_down_cb = SARSimTouchDownCB;
            realm->collision_cb_client_data = core_ptr;
            realm->collision_cb = SARSimCollisionCB;

	}

	/* Load textures specified in core structure's textures list
	 * first before loading textures specified in the scene file.
	 */
	if(1)
	{
	    int i, n, tex_fmt;
            char *name, *path;
	    sar_texture_name_struct *tn_ptr;
            char tmp_path[PATH_MAX + NAME_MAX];

/* Already reset.
	    scene->total_texture_refs = 0;
	    scene->texture_ref = NULL;
 */

	    tex_fmt = V3D_TEX_FORMAT_RGBA;

	    /* Go through texture referance names list on core
	     * structure.
	     */
	    for(i = 0; i < core_ptr->total_texture_list; i++)
	    {
		tn_ptr = core_ptr->texture_list[i];
		if(tn_ptr == NULL)
		    continue;

		name = tn_ptr->name;
		path = tn_ptr->filename;

		if((name == NULL) ||
                   (path == NULL)
		)
		    continue;

		/* Allocate a new texture. */
		n = scene->total_texture_refs;
		scene->total_texture_refs++;
		scene->texture_ref = (v3d_texture_ref_struct **)realloc(
		    scene->texture_ref,
		    scene->total_texture_refs * sizeof(v3d_texture_ref_struct *)
		);
		if(scene->texture_ref == NULL)
		{
		    scene->total_texture_refs = 0;
		    break;
		}

		strptr = PrefixPaths(dname.local_data, path);
		if(strptr != NULL)
		{
		    if(stat(strptr, &stat_buf))
			strptr = PrefixPaths(dname.global_data, path);
		}
		if(strptr != NULL)
		{
		    strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
		    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
		    scene->texture_ref[n] = V3DTextureLoadFromFile2D(
			tmp_path, name,
			tex_fmt,
			NULL, NULL
		    );
		}
	    }
	}

	/* Load referances to default sound objects. */
	if(1)
	{
            char *name, *storage;

#define SAR_DO_LOAD_SOUND	\
{ \
 if(name == NULL) \
 { \
  storage = NULL; \
 } \
 else \
 { \
  strptr = PrefixPaths(dname.local_data, name); \
  if(strptr != NULL) \
  { \
   if(stat(strptr, &stat_buf)) \
    strptr = PrefixPaths(dname.global_data, name); \
  } \
  storage = StringCopyAlloc(strptr); \
 } \
}
	    name = SAR_DEF_SOUND_LAND_WHEEL_SKID;
	    SAR_DO_LOAD_SOUND
	    scene->land_wheel_skid_sndpath = storage;

            name = SAR_DEF_SOUND_LAND_SKI_SKID;
            SAR_DO_LOAD_SOUND
            scene->land_ski_skid_sndpath = storage;

            name = SAR_DEF_SOUND_LAND_SKI;
            SAR_DO_LOAD_SOUND
            scene->land_ski_sndpath = storage;

            name = SAR_DEF_SOUND_LAND_BELLY;
            SAR_DO_LOAD_SOUND
            scene->land_belly_sndpath = storage;

            name = SAR_DEF_SOUND_CRASH_OBSTRUCTION;
            SAR_DO_LOAD_SOUND
            scene->crash_obstruction_sndpath = storage;

            name = SAR_DEF_SOUND_CRASH_GROUND;
            SAR_DO_LOAD_SOUND
            scene->crash_ground_sndpath = storage;

#undef SAR_DO_LOAD_SOUND
	}

	/* Create horizon. */
	horizon_ptr = &scene->horizon;
        horizon_ptr->last_tod = -1;
        horizon_ptr->texture = NULL;
        horizon_ptr->total_textures = 0;


	/* Generate default weather settings. */
	SARWeatherSetScenePreset(
	    core_ptr->weather_data,
	    scene,
	    weather_preset_name
	);


        /* Begin reading scene file. */
        while(1)
        {
	    if(progress_func != NULL)
	    {
		file_pos = ftell(fp);

		if(progress_func(client_data, file_pos, file_size))
		    break;
	    }

            c = fgetc(fp);
            if(c == EOF) 
                break;  
                     
            /* Read past leading spaces. */
            if(ISBLANK(c))
            {
                FSeekPastSpaces(fp);
         
                /* Get first non space character. */
                c = fgetc(fp);
                if(c == EOF)
                    break;  
            }

            lines_read++;

            /* Newline character? */
            if(ISCR(c))
                continue;


            /* Comment? */
            if(ISCOMMENT(c))
            {
                FSeekNextLine(fp);        /* Seek to next line. */
                continue;
            }


            /* Seek back one character and fetch as operation. */
            fseek(fp, -1, SEEK_CUR);

            /* Get pointer to operation string. */
            strptr = SARObjFSeekGetOpString(fp);


	    /* Scene name. */
	    if(!strcasecmp(strptr, "name"))
            {
		FSeekNextLine(fp);
	    }
	    /* Register location. */
	    else if(!strcasecmp(strptr, "register_location"))
	    {
		FSeekNextLine(fp);
	    }
	    /* Entire scene map. */
	    else if(!strcasecmp(strptr, "entire_scene_map"))
            {
               FSeekNextLine(fp);
            }
            /* Create new object. */   
            else if(!strcasecmp(strptr, "create_object"))
	    {
                FGetValuesF(fp, value, 1);

                obj_num = SARObjCreate(
		    scene, ptr, total,
		    (int)value[0]
		);
		obj_ptr = ((obj_num < 0) ? NULL : (*ptr)[obj_num]);

		/* Reset all substructure type pointers. */
		DO_RESET_SUBSTRUCTURE_PTRS

		if(obj_ptr != NULL)
		{
		    /* Get pointer to substructure. */
		    switch(obj_ptr->type)
		    {
		      case SAR_OBJ_TYPE_AIRCRAFT:
		        obj_aircraft_ptr = obj_ptr->data;
		        break;

		      case SAR_OBJ_TYPE_GROUND:
		        obj_ground_ptr = obj_ptr->data;
		        break;
		    }
		}
	    }
	    /* Create runway. */
            else if(!strcasecmp(strptr, "create_runway"))
            {
                FGetValuesF(fp, value, 3);

                obj_num = SARObjCreate( 
                    scene, ptr, total,  
                    SAR_OBJ_TYPE_RUNWAY
                );
                obj_ptr = ((obj_num < 0) ? NULL : (*ptr)[obj_num]);

                if(obj_ptr != NULL)
                {
                    sar_object_runway_struct *obj_runway_ptr = obj_ptr->data;

                    obj_ptr->range = value[0];

                    if(obj_runway_ptr != NULL)
                    {
                        obj_runway_ptr->flags |= SAR_RUNWAY_FLAG_NORTH_LABEL;
                        obj_runway_ptr->flags |= SAR_RUNWAY_FLAG_SOUTH_LABEL;
                        obj_runway_ptr->flags |= SAR_RUNWAY_FLAG_EDGE_LIGHTING;
                        obj_runway_ptr->flags |= SAR_RUNWAY_FLAG_NORTH_LIGHTING;
                        obj_runway_ptr->flags |= SAR_RUNWAY_FLAG_SOUTH_LIGHTING;

                        obj_runway_ptr->length = value[1];
                        obj_runway_ptr->width = value[2];

                        SARObjAddContactBoundsRectangular(
                            obj_ptr,
                            SAR_CRASH_FLAG_SUPPORT_SURFACE, 0,
			    -(obj_runway_ptr->width / 2),
			    (obj_runway_ptr->width / 2),
			    -(obj_runway_ptr->length / 2),
                            (obj_runway_ptr->length / 2),
                            0.0, 0.0
			);

/*
                        obj_runway_ptr->north_label =
                        obj_runway_ptr->south_label =
 */

                        obj_runway_ptr->light_spacing =
                            obj_runway_ptr->length / 50;

                        /* Load texture. */
                        obj_runway_ptr->tex_num = SARGetTextureRefNumberByName(
                            scene, SAR_STD_TEXNAME_RUNWAY
                        );
                    }
                }
            }
	    /* Create helipad. */
            else if(!strcasecmp(strptr, "create_helipad"))
            {
                FGetValuesF(fp, value, 3);

                obj_num = SARObjCreate(
                    scene, ptr, total,
                    SAR_OBJ_TYPE_HELIPAD
                );
                obj_ptr = ((obj_num < 0) ? NULL : (*ptr)[obj_num]);

		if(obj_ptr != NULL)
		{
		    sar_object_helipad_struct *obj_helipad_ptr = obj_ptr->data;

		    obj_ptr->range = value[0];

		    if(obj_helipad_ptr != NULL)
		    {
			obj_helipad_ptr->flags |= SAR_HELIPAD_FLAG_LABEL;
                        obj_helipad_ptr->flags |= SAR_HELIPAD_FLAG_EDGE_LIGHTING;

                        obj_helipad_ptr->length = value[1];
                        obj_helipad_ptr->width = value[2];

                        SARObjAddContactBoundsRectangular(
                            obj_ptr,
                            SAR_CRASH_FLAG_SUPPORT_SURFACE, 0,
                            -(obj_helipad_ptr->width / 2),
                            (obj_helipad_ptr->width / 2),
                            -(obj_helipad_ptr->length / 2),
                            (obj_helipad_ptr->length / 2),
                            0.0, 0.0
                        );

			obj_helipad_ptr->label = StringCopyAlloc("PVT");

			obj_helipad_ptr->light_spacing =
			    obj_helipad_ptr->length / 10;

			/* Load texture. */
                        obj_helipad_ptr->tex_num = SARGetTextureRefNumberByName(
			    scene, SAR_STD_TEXNAME_HELIPAD
			);
		    }
		}
	    }
	    /* Create human. */
	    else if(!strcasecmp(strptr, "create_human"))
            {
                FGetValuesF(fp, value, 6);

		/* Create human object. */
                obj_num = SARObjCreate(
                    scene, ptr, total,
                    SAR_OBJ_TYPE_HUMAN
                );
                obj_ptr = ((obj_num < 0) ? NULL : (*ptr)[obj_num]);

                if(obj_ptr != NULL)
                {
                    sar_object_human_struct *obj_human_ptr = obj_ptr->data;

                    obj_ptr->range = 400;

		    SARObjAddContactBoundsCylendrical(
			obj_ptr,
			0, 0,		/* Crash flags, crash type. */
			SAR_HUMAN_CONTACT_RADIUS,
			SAR_HUMAN_CONTACT_ZMIN, SAR_HUMAN_CONTACT_ZMAX
		    );

		    obj_human_ptr->flags = 0;

		    /* Need rescue? */
		    if((int)value[0])
			obj_human_ptr->flags |= SAR_HUMAN_FLAG_NEED_RESCUE;

                    /* Sitting? */
                    if((int)value[1])
                        obj_human_ptr->flags |= SAR_HUMAN_FLAG_SIT;

                    /* Lying? */
                    if((int)value[2])
                        obj_human_ptr->flags |= SAR_HUMAN_FLAG_LYING;

                    /* Alert? */
                    if((int)value[3])
                        obj_human_ptr->flags |= SAR_HUMAN_FLAG_ALERT;

                    /* Aware? */
                    if((int)value[4])
                        obj_human_ptr->flags |= SAR_HUMAN_FLAG_AWARE;

		    /* In water? */
		    if((int)value[5])
			obj_human_ptr->flags |= SAR_HUMAN_FLAG_IN_WATER;

		    obj_human_ptr->anim_rate = SAR_HUMAN_ANIM_RATE;

		    /* Water ripples texture. */
		    obj_human_ptr->water_ripple_tex_num =
			SARGetTextureRefNumberByName(scene, SAR_STD_TEXNAME_WATER_RIPPLE);

                    obj_human_ptr->intercepting_object = -1;
		}
            }
            /* Create permodeled. */
            else if(!strcasecmp(strptr, "create_premodeled"))
            {
                strptr2 = FGetString(fp);
                if(strptr2 != NULL)
		{
		    argv = strexp(strptr2, &argc);

		    obj_num = SARObjCreatePremodeled(
			core_ptr, scene, argc, argv
		    );
                    obj_ptr = ((obj_num < 0) ? NULL : (*ptr)[obj_num]);

		    StringFreeArray(argv, argc);
		}
		free(strptr2);
		strptr2 = NULL;
	    }

            /* Object name. */
            else if(!strcasecmp(strptr, "object_name"))
            {
                strptr2 = FGetString(fp);
                if(obj_ptr != NULL) 
                {
                    free(obj_ptr->name);
                    obj_ptr->name = strptr2;
		    strptr2 = NULL;
                }
		free(strptr2);
            }

            /* Texture base directory. */
            else if(!strcasecmp(strptr, "texture_base_directory") ||
                    !strcasecmp(strptr, "texture_base_dir")   
            )
            {
                FSeekNextLine(fp);  /* Skip this. */
            }
            /* Texture load. */
            else if(!strcasecmp(strptr, "texture_load"))
            {
		strptr2 = FGetString(fp);
                SARObjLoadLineTexture(
		    core_ptr, scene,
		    strptr2
		);
		free(strptr2);
            }
            /* Model file. */
            else if(!strcasecmp(strptr, "model_file"))
            {
                strptr2 = FGetString(fp);
                if((strptr2 != NULL) && (obj_ptr != NULL))
                    SARObjLoadFromFile(core_ptr, obj_num, strptr2);

		free(strptr2);
            }
            /* Translation (overrides value specified in object file). */  
            else if(!strcasecmp(strptr, "translate") ||
                    !strcasecmp(strptr, "translation")
	    )
            {
		sar_position_struct new_pos;

                FGetValuesF(fp, value, 3);

                new_pos.x = value[0];
                new_pos.y = value[1];
                new_pos.z = SFMFeetToMeters(value[2]);

		if(obj_ptr != NULL)
		    SARSimWarpObject(scene, obj_ptr, &new_pos, NULL);
            }
            /* Rotate (overrides value specified in object file). */
            else if(!strcasecmp(strptr, "rotate"))
            {
		sar_direction_struct new_dir;

                FGetValuesF(fp, value, 3);

                /* Values are in degrees. */
                new_dir.heading = SFMDegreesToRadians(value[0]);
                new_dir.pitch = SFMDegreesToRadians(value[1]);
                new_dir.bank = SFMDegreesToRadians(value[2]);

                if(obj_ptr != NULL)
                    SARSimWarpObject(scene, obj_ptr, NULL, &new_dir);
	    }
            /* Visable range. */
            else if(!strcasecmp(strptr, "range"))
            {
                FGetValuesF(fp, value, 1);
                if(obj_ptr != NULL)  
                {
                    obj_ptr->range = value[0];

                    if(value[0] < 0.0)
                        fprintf(stderr,
 "%s: Line %i: Warning: range is negative.\n",
                            filename, lines_read
                        );
                }
	    }
            /* No depth test? */
            else if(!strcasecmp(strptr, "no_depth_test"))
            {
                FSeekNextLine(fp);

		if(obj_ptr != NULL)
		{
                    obj_ptr->flags |= SAR_OBJ_FLAG_NO_DEPTH_TEST;
                }
	    }

            /* Contact bounds, spherical shape. */
            else if(!strcasecmp(strptr, "contact_spherical"))
            {
                sar_contact_bounds_struct *cb = obj_ptr->contact_bounds;

                /* Arguments are: <radius>
                 */
		FGetValuesF(fp, value, 1);

                SARObjAddContactBoundsSpherical(
                    obj_ptr,
                    (cb == NULL) ? 0 : cb->crash_flags,
                    (cb == NULL) ? 0 : cb->crash_type,
                    value[0]                    /* Radius. */
                );
            }
            /* Contact bounds, cylendrical shape. */
            else if(!strcasecmp(strptr, "contact_cylendrical"))
            {
                sar_contact_bounds_struct *cb = obj_ptr->contact_bounds;

                /* Arguments are: <radius> <height_min> <height_max>   
                 */
                FGetValuesF(fp, value, 3); 

                SARObjAddContactBoundsCylendrical( 
                    obj_ptr,
                    (cb == NULL) ? 0 : cb->crash_flags,
                    (cb == NULL) ? 0 : cb->crash_type,
                    value[0],                   /* Radius. */
                    value[1], value[2]          /* Height min/max. */
                );
            }
            /* Contact bounds, rectangular shape. */  
            else if(!strcasecmp(strptr, "contact_rectangular"))
            {
                sar_contact_bounds_struct *cb = obj_ptr->contact_bounds;

                /* Arguments are: <x_min> <x_max> <y_min> <y_max>
                 * <z_min> <z_max>
                 */
                FGetValuesF(fp, value, 6);

                SARObjAddContactBoundsRectangular(
                    obj_ptr,
                    (cb == NULL) ? 0 : cb->crash_flags,
                    (cb == NULL) ? 0 : cb->crash_type,
                    value[0], value[1],         /* X min/max. */
                    value[2], value[3],         /* Y min/max. */
                    value[4], value[5]          /* Z min/max. */
                );
            }

            /* Fuel (overrides value specified in object file). */
            else if(!strcasecmp(strptr, "fuel"))
            {
                FGetValuesF(fp, value, 2);

                if((value[0] < 0) || (value[1] < 0))
                    fprintf(stderr,
 "%s: Line %i: Warning: fuel value is negative.\n",
                        filename, lines_read
                    );

                if(obj_aircraft_ptr != NULL)
                {
                    obj_aircraft_ptr->fuel = value[0];
                    obj_aircraft_ptr->fuel_max = value[1];
                }
            }

            /* Ground elevation. */
            else if(!strcasecmp(strptr, "ground_elevation"))
            {
                FGetValuesF(fp, value, 1);

		if(obj_ground_ptr == NULL)
		{
                    fprintf(stderr,
 "%s: Line %i: ground_elevation: Object not of type `%i'.\n",
                        filename, lines_read, SAR_OBJ_TYPE_GROUND
                    );
		}
		else
		{
                    if(value[0] < 0)
                        fprintf(stderr,
 "%s: Line %i: Warning: ground_elevation value is negative.\n",
                            filename, lines_read
                        );

                    obj_ground_ptr->elevation = SFMFeetToMeters(value[0]);
                }
	    }

	    /* Set human rescue message. */
            else if(!strcasecmp(strptr, "set_human_message_hoist_in"))
            {
		strptr = FGetString(fp);
                if((strptr != NULL) && (obj_ptr != NULL))
		{
                    if(obj_ptr->type == SAR_OBJ_TYPE_HUMAN)
                    {
                        sar_object_human_struct *obj_human_ptr = obj_ptr->data;

			if(obj_human_ptr != NULL)
			{
			    free(obj_human_ptr->mesg_hoist_in);
			    obj_human_ptr->mesg_hoist_in = StringCopyAlloc(strptr);
			}
                    }
		    else
		    {
			fprintf(
			    stderr,
 "%s: Line %i: set_human_message_hoist_in: Object not of type %i.\n",
			    filename, lines_read, SAR_OBJ_TYPE_HUMAN
			);
		    }
                }
		free(strptr);
	    }


	    /* Global scene elevation from MSL. */
	    else if(!strcasecmp(strptr, "scene_msl_elevation"))
	    {
                FGetValuesF(fp, value, 1);

		scene->msl_elevation = SFMFeetToMeters(value[0]);
	    }
	    /* Global scene planet units. */
            else if(!strcasecmp(strptr, "scene_planet_units"))
            {
		FGetValuesF(fp, value, 3);

                scene->dms_x_offset = value[0];
		scene->dms_y_offset = value[1];
		scene->planet_radius = value[2];
	    }
	    /* Global scene light source position. */
	    else if(!strcasecmp(strptr, "scene_light_position"))
            {
/*
		FGetValuesF(fp, value, 3);

		scene->light_pos.x = value[0];
                scene->light_pos.y = value[1];
                scene->light_pos.z = SFMFeetToMeters(value[2]);
 */
	    }
            /* Global scene light source color. */
            else if(!strcasecmp(strptr, "scene_light_color"))
            {
/*
                FGetValuesF(fp, value, 4);

                scene->light_color.r = value[0];
                scene->light_color.g = value[1];
		scene->light_color.b = value[2];
		scene->light_color.a = value[3];
 */
            }
	    /* Global scene ground base tile. */
	    else if(!strcasecmp(strptr, "scene_ground_tile"))
            {
                double min, max;

		GLint list;
                sar_visual_model_struct **vmodel;
		v3d_texture_ref_struct *t = NULL;


		strptr2 = FGetString(fp);
		if(strptr2 != NULL)
		{
		    argv = strexp(strptr2, &argc);
		    free(strptr2);
		}
		else
		{
		    argv = NULL;
		    argc = 0;
		}

                /* Args 0 to 2: Set tile size and close range in meters. */
                scene->base.tile_width = ((argc > 0) ?
                    atoi(argv[0]) : DEF_GROUND_BASE_TILE_WIDTH
                );
                scene->base.tile_height = ((argc > 1) ?
                    atoi(argv[1]) : DEF_GROUND_BASE_TILE_HEIGHT
                );
                scene->base.close_range = ((argc > 2) ?
                    atof(argv[2]) : DEF_GROUND_BASE_TILE_CLOSE_RANGE
                );

                /* Args 3 to 6: Solid color of ground base. */
                scene->base.color.r = ((argc > 3) ? atof(argv[3]) : 1.0);
                scene->base.color.g = ((argc > 4) ? atof(argv[4]) : 1.0);
                scene->base.color.b = ((argc > 5) ? atof(argv[5]) : 1.0);
                scene->base.color.a = ((argc > 6) ? atof(argv[6]) : 1.0);

                /* Arg 7: Get referance to texture. */
                t = ((argc > 7) ?
                    SARGetTextureRefByName(scene, argv[7]) : NULL
                );

		/* Create simple (far away) ground base visual model as
		 * needed.
		 */
		vmodel = &scene->base.visual_model_simple;
		if((*vmodel) != NULL)
		{
                    fprintf(
                        stderr,
 "%s: Line %i: Warning: Ground base (far away) visual model is already defined.\n",
                        filename, lines_read
		    );
		}
		else
		{
		    /* Create new visual model. */
		    (*vmodel) = SARVisualModelNew(
			scene, NULL, NULL
		    );
		    /* Increment referance count so it should now be 1. */
		    SARVisualModelRef(*vmodel);
		    /* (Re)generate GL list on visual model structure. */
		    list = (GLint)SARVisualModelNewList(*vmodel);
		    if(list != 0)
		    {
			/* Mark visual model as loading. */
			(*vmodel)->load_state = SAR_VISUAL_MODEL_LOADING;

			/* Begin recording new list. */
			glNewList(list, GL_COMPILE);
			{
			    /* Set shading type. */
			    glShadeModel(GL_FLAT);

			    /* Unselect texture. */
/*
			    V3DTextureSelect(NULL);
			    tex_orient = TEX_ORIENT_NONE;
 */
			    /* Far (big) ground base tiles. */
			    min = -(SAR_MAX_VISIBILITY_DISTANCE +
				(0.25 * SAR_MAX_VISIBILITY_DISTANCE));
			    max = SAR_MAX_VISIBILITY_DISTANCE +
				(0.25 * SAR_MAX_VISIBILITY_DISTANCE);
			    SARObjGenerateTilePlane(
				min, max,		/* Min and max. */
				(int)(max / 4),		/* Tile width and height. */
				(int)(max / 4)
			    );
			}
			glEndList();

                        /* Mark visual model as done loading. */
                        (*vmodel)->load_state = SAR_VISUAL_MODEL_LOADED;
		    }
		}

                /* Create detailed (close up) ground base visual model as
                 * needed.
                 */
                vmodel = &scene->base.visual_model_close;
                if((*vmodel) != NULL)
                {
                    fprintf(
                        stderr,
 "%s: Line %i: Warning: Ground base (close up) visual model is already defined.\n",
                        filename, lines_read
                    );
                }
                else
                {
                    /* Create new visual model. */
                    (*vmodel) = SARVisualModelNew(
                        scene, NULL, NULL
                    );
                    /* Increment referance count so it should now be 1. */
                    SARVisualModelRef(*vmodel);
                    /* (Re)generate GL list on visual model structure. */
                    list = (GLint)SARVisualModelNewList(*vmodel);
                    if(list != 0)
                    {
                        /* Mark visual model as loading. */
                        (*vmodel)->load_state = SAR_VISUAL_MODEL_LOADING;

                        /* Begin recording new list. */
                        glNewList(list, GL_COMPILE);
                        {
                            /* Set shading type. */
                            glShadeModel(GL_FLAT);

			    /* Select texture if defined. */
			    V3DTextureSelect(t);

			    /* Close range textured tiles. */
			    SARObjGenerateTilePlane(
				-scene->base.close_range,	/* Min. */
				scene->base.close_range,	/* Max. */
				scene->base.tile_width,		/* Tile width. */
				scene->base.tile_height		/* Tile height. */
			    );
			}
			glEndList();

                        /* Mark visual model as done loading. */
                        (*vmodel)->load_state = SAR_VISUAL_MODEL_LOADED;
		    }
		}

		StringFreeArray(argv, argc);
            }
            /* Unknown operation. */
            else
            {
                fprintf(
                    stderr,
                    "%s: Line %i: Unsupported operation `%s'\n",
                    filename, lines_read, strptr
                );
		FSeekNextLine(fp);        /* Seek to next line. */
            }
	}

	/* Close the file. */
        FClose(fp);


        return(0);
}
