#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "sfm.h"

#include "horizon.h"
#include "sarreality.h"

#include "obj.h"

#include "simcb.h"
#include "simutils.h"
#include "simsurface.h"
#include "simcontact.h"
#include "simop.h"
#include "simmanage.h"

#include "sar.h"


int SARSimUpdateScene(
        sar_core_struct *core_ptr,
        sar_scene_struct *scene
);
int SARSimUpdateSceneObjects(
        sar_core_struct *core_ptr,
        sar_scene_struct *scene
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))


/*
 *	Updates the given scene structure's members (but does not update
 *	any objects).
 *
 *	Time of day will be updated on the scene, then along with the
 *	primary light position light_pos and color light_color.
 *
 *	Horizon texture will be (re)generated as need per a given interval
 *	(every 5 minutes).
 *
 *	Returns 0 on success and non-zero on error or memory pointer
 *	change.
 */
int SARSimUpdateScene(sar_core_struct *core_ptr, sar_scene_struct *scene)
{
	double time_coeff;
	int new_tod_code;
	sar_position_struct *cam_pos, *pos;
	sar_color_struct *c;
	sar_scene_horizon_struct *horizon_ptr;

/* Macro to convert hours into seconds. */
#define HTOS(h)	((h) * 3600.0)

	if((core_ptr == NULL) ||
           (scene == NULL)
	)
	    return(-1);

	horizon_ptr = &scene->horizon;
	cam_pos = &scene->camera_pos;

	/* Update time of day (in seconds) since midnight. */
	scene->tod += (double)(
	    (double)lapsed_millitime * time_compression / 1000
	);

	/* Sanitize time of day. */
	while(scene->tod > HTOS(24.0))
	    scene->tod -= (double)HTOS(24.0);


	/* Update time of day code depending on time of day. */
	if(scene->tod < HTOS(6.0))		/* Before 6:00 am. */
	    new_tod_code = SARSceneTODCodeNight;
	else if(scene->tod < HTOS(8.0))		/* Before 8:00 am. */
	    new_tod_code = SARSceneTODCodeDawn;
        else if(scene->tod < HTOS(19.0))	/* Before 7:00 pm. */
            new_tod_code = SARSceneTODCodeDay;
	else if(scene->tod < HTOS(20.0))	/* Before 8:00 pm. */
	    new_tod_code = SARSceneTODCodeDusk;
	else					/* After 8:00 pm. */
	    new_tod_code = SARSceneTODCodeNight;

	scene->tod_code = new_tod_code;


	/* Move primary light position relative to camera based
	 * on time of day and also change primary light color.
	 */
	pos = &scene->light_pos;
	c = &scene->light_color;
	if(scene->tod > (12 * 3600))
	{
	    double g;

	    /* Afternoon or later. */
	    time_coeff = 1.0 - MIN(
		(scene->tod - (12 * 3600)) / (9 * 3600),
		1.0
	    );

	    pos->x = cam_pos->x - (
		SAR_PRILIGHT_TO_CAMERA_RAD * (1.0 - time_coeff)
	    );
	    pos->y = cam_pos->y;
	    pos->z = 0.0 + (
		SAR_PRILIGHT_TO_CAMERA_RAD * time_coeff
	    );

	    g = pow(time_coeff, 0.3);	/* Concave down light curve. */
	    c->a = 1.0;
	    c->r = g;
            c->g = g;
            c->b = g;
	}
	else
	{
	    double g;

	    /* Morning. */
            time_coeff = MAX(
                (scene->tod - (6 * 3600)) / (6 * 3600),
                0.0
            );

            pos->x = cam_pos->x + (
                SAR_PRILIGHT_TO_CAMERA_RAD * (1.0 - time_coeff)
            );
            pos->y = cam_pos->y;
            pos->z = 0.0 + (
                SAR_PRILIGHT_TO_CAMERA_RAD * time_coeff
            );

            g = pow(time_coeff, 0.3);	/* Concave down light curve. */
            c->a = 1.0;
            c->r = g;
            c->g = g;
            c->b = g;
	}


	/* Regenerate horizon as needed every 5 minutes. */
	if(horizon_ptr->last_tod != (int)(scene->tod / (5 * 60)))
	{
	    int i, total, tex_num;
	    double rc, gc, bc;		/* Coefficient color change. */
	    double darken_coeff, midpoint = 0.88;
	    sar_color_struct start_color, end_color;
	    /* Lumination coeff to 6:00 and 18:00. 0.0 is darkest and
	     * 1.0 is brightest.
	     */
	    double lum_coeff = (scene->tod - (12 * 3600)) / (6 * 3600);

	    /* Lumination coeff must be positive. */
	    if(lum_coeff < 0.0)
		lum_coeff *= -1.0;
	    if(lum_coeff > 1.0)
		lum_coeff = 1.0;
	    lum_coeff = 1.0 - lum_coeff;	/* Flip. */


	    /* Update last tod to current time in 5 minute units. */
	    horizon_ptr->last_tod = (int)(scene->tod / (5 * 60));

/* Creates a new horizon texture on the horizon_ptr. */
#define DO_CREATE_TEXTURE	\
{ \
 tex_num = horizon_ptr->total_textures; \
 horizon_ptr->total_textures++; \
 horizon_ptr->texture = (v3d_texture_ref_struct **)realloc( \
  horizon_ptr->texture, \
  horizon_ptr->total_textures * sizeof(v3d_texture_ref_struct *) \
 ); \
 if(horizon_ptr->texture == NULL) \
 { \
  horizon_ptr->total_textures = 0; \
 } \
 else \
 { \
  horizon_ptr->texture[tex_num] = SARCreateHorizonTexture( \
   NULL, /* No name. */ \
   &start_color, &end_color, \
   32, midpoint \
  ); \
 } \
}

	    /* Adjust gradient horizon textures by recreating them.
	     * Do not add lumination/gamma to them, the light color
	     * will be multiplied during drawing.
	     */

	    for(i = 0; i < horizon_ptr->total_textures; i++)
		V3DTextureDestroy(horizon_ptr->texture[i]);
	    free(horizon_ptr->texture);
	    horizon_ptr->texture = NULL;
	    horizon_ptr->total_textures = 0;


	    /* Create textures, starting with most brightest. */
	    memcpy(
		&start_color, &scene->sky_nominal_color,
		sizeof(sar_color_struct)
	    );

	    total = 5;
	    for(i = 0; i < total; i++)
	    {
		darken_coeff = (double)i / (double)(total - 1);

		rc = MIN(
		    (scene->sky_brighten_color.r * (1.0 - darken_coeff)) +
		    (scene->sky_darken_color.r * darken_coeff),
		    1.0
		);
                gc = MIN(
                    (scene->sky_brighten_color.g * (1.0 - darken_coeff)) +
                    (scene->sky_darken_color.g * darken_coeff),
                    1.0
                );
                bc = MIN(
                    (scene->sky_brighten_color.b * (1.0 - darken_coeff)) +
                    (scene->sky_darken_color.b * darken_coeff),
                    1.0
                );

                end_color.r = MIN(
		    (1.0 * lum_coeff) + (rc * (1.0 - lum_coeff)),
		    1.0
		);
                end_color.g = MIN(
                    (1.0 * lum_coeff) + (gc * (1.0 - lum_coeff)), 
                    1.0 
                );
                end_color.b = MIN(
                    (1.0 * lum_coeff) + (bc * (1.0 - lum_coeff)), 
                    1.0 
                );

		DO_CREATE_TEXTURE
	    }
	}
#undef DO_CREATE_TEXTURE
#undef HTOS

	return(0);
}

/*
 *	First updates the timmings on the scene structure's SFMRealmStruct
 *	and then updates all objects with respect to the given scene and
 *	core structure.
 *
 *	Returns 0 on success and non-zero on error or memory pointer
 *	change.
 */
int SARSimUpdateSceneObjects(
	sar_core_struct *core_ptr, sar_scene_struct *scene
)
{
	int i, status;
        int spawn_smoke = 0;
	int *obj_total;
        sar_object_struct *obj_ptr, ***obj_pa;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_contact_bounds_struct *cb;
	SFMModelStruct *fdm;


	if((core_ptr == NULL) ||
           (scene == NULL)
	)
	    return(-1);

	/* Time to spawn smoke? */
	if(next.spawn_smoke <= cur_millitime)
	    spawn_smoke = 1;

	/* Update timming on realm. */
	if(scene->realm != NULL)
	{
	    SFMSetTimming(scene->realm, lapsed_millitime);
	    SFMSetTimeCompression(scene->realm, time_compression);
	}


	/* Begin handling each object on core structure. */
        obj_pa = &core_ptr->object;
        obj_total = &core_ptr->total_objects;
        for(i = 0; i < (*obj_total); i++)
        {
            obj_ptr = (*obj_pa)[i];
            if(obj_ptr == NULL)
                continue;

	    status = 0;		/* Reset `SFM handled' value. */

	    /* SFM realm structure allocated? */
	    if(scene->realm != NULL)
	    {
		/* Handle by object type. */
		switch(obj_ptr->type)
		{
		  case SAR_OBJ_TYPE_AIRCRAFT:
		    obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
		    if(obj_aircraft_ptr == NULL)
			break;

                    fdm = obj_aircraft_ptr->fdm;
                    if(fdm == NULL)
			break;

		    /* Set object values to SFM structure. */
                    SARSimSetSFMValues(scene, obj_ptr);

                    /* Apply control positions if this is the
                     * player object.
                     */
                    if(obj_ptr == scene->player_obj_ptr)
		    {
                        SARSimApplyGCTL(core_ptr, obj_ptr);
		    }
		    else
		    {

 /* Otherwise apply AI? */

		    }

		    /* Set status to 1, noting that the SFM structure
		     * was handled (which we will actually do so just
		     * below).
		     */
		    status = 1;

                    /* Begin SFM updating. */
                    if(SFMForceApplyNatural(scene->realm, fdm))
			break;
		    if(SFMForceApplyArtificial(scene->realm, fdm))
			break;
		    if(SFMForceApplyControl(scene->realm, fdm))
			break;

                    /* Get new object values from processed SFM. */
                    SARSimGetSFMValues(scene, obj_ptr);
		    break;

 /* Add SFM handling support for other types of objects as needed here. */

		}
	    }
	    /* Check if object survived SFM handling, in other words is
	     * it still allocated?
	     */
	    if(status)
	    {
		/* SFM was handled on the object, so check if object
		 * is still allocated.
		 */
		if(SARObjIsAllocated(*obj_pa, *obj_total, i))
		    obj_ptr = (*obj_pa)[i];
		else
		    continue;
	    }

            /* Apply natural forces to object. */
            SARSimApplyNaturalForce(core_ptr, obj_ptr);

            /* Apply artificial forces to object. */
            SARSimApplyArtificialForce(core_ptr, obj_ptr);

            /* Get pointer to object's contact bounds structure (which
	     * may be NULL).
	     */
	    cb = obj_ptr->contact_bounds;

	    /* Can this object crash into other objects? */
            if(((cb == NULL) ? 0 : cb->crash_flags) &
		SAR_CRASH_FLAG_CRASH_OTHER
	    )
            {
		/* Perform object to object crash contact check. */
                if(SARSimCrashContactCheck(core_ptr, obj_ptr) > -1)
                {
                    /* Crash contact into another object has occured,
		     * so we need to check if this object is still
                     * allocated.
                     */
                    if(SARObjIsAllocated(*obj_pa, *obj_total, i))
                        obj_ptr = (*obj_pa)[i];
                    else
                        continue;
                }
            }

            /* Rescue basket contact check (only if this object is a
	     * type that has a rescue basket). If a contact is made
	     * then a valid index is returned and the proper procedure
	     * will have been taken.
	     */
	    if(SARSimRescueBasketContactCheck(core_ptr, obj_ptr) > -1)
	    {
		/* Rescue basket contact has occured, need to check if
		 * this object is still allocated.
		 */
		if(SARObjIsAllocated(*obj_pa, *obj_total, i))
		    obj_ptr = (*obj_pa)[i];
		else
		    continue;
            }

            /* Spawn smoke for object as needed. */
            if(spawn_smoke)
                SARSimSmokeSpawnCheck(
                    core_ptr,  
                    obj_ptr,
                    cur_millitime + option.smoke_life_span  
                );

            /* Check if the object is `too old', if it is then it
	     * will be deleted.
	     */
            if(SARSimDoMortality(core_ptr, obj_ptr))
	    {
                /* Object has been deleted, check if it is still
		 * allocated (which it probably won't be).
                 */
                if(SARObjIsAllocated(*obj_pa, *obj_total, i))
                    obj_ptr = (*obj_pa)[i];
		else
		    continue;
	    }

/* Add more object updates here. */

        }

	return(0);
}
